import psutil
import dask
import numpy as np
import argparse
import traceback
import time
import sys
import os
import logging
from casatasks import casalog
from dask import delayed
from dask.distributed import Client
from meersolar.utils import *

logging.getLogger("distributed").setLevel(logging.WARNING)


try:
    casalogfile = casalog.logfile()
    os.system("rm -rf " + casalogfile)
except BaseException:
    pass


datadir = get_datadir()


def get_polmodel_coeff(model_file):
    """
    Get polarization model coefficients

    Parameters
    ----------
    model_file : str
        Model ascii file name

    Returns
    -------
    float
        Reference frequency in GHz
    float
        Stokes I flux density in Jy at reference frequency
    list
        Stokes I spectrum log-polynomial coefficients
    list
        List of linear polarization fraction polynomial coefficients
    list
        List of linear polarization angle (in radians) coefficients
    """
    freq, I, pfrac, pangle = np.loadtxt(model_file, unpack=True, skiprows=5)
    ref_freq = freq[0]
    x = (freq - ref_freq) / ref_freq
    polyI = np.polyfit(np.log10(x[1:]), np.log10(I[1:]), deg=5)[::-1]
    poly_pfrac = np.polyfit(x, pfrac, deg=5)[::-1]
    poly_pangle = np.polyfit(x, pangle, deg=5)[::-1]
    return ref_freq, I[0], polyI[1:], poly_pfrac, poly_pangle


def polcal_setjy(msname="", field_name="", n_threads=-1, ismms=True, dry_run=False):
    """
    Setjy polcal fields (3C286 or 3C138)

    Parameters
    ----------
    msname : str
        Name of measurement set
    field : str, optional
        Field name
    n_threads : int, optional
        Number of OpenMP threads
    """
    limit_threads(n_threads=n_threads)
    from casatasks import setjy

    if dry_run:
        process = psutil.Process(os.getpid())
        mem = round(process.memory_info().rss / 1024**3, 2)  # in GB
        return mem
    try:
        print(f"Polarization calibrator field name : {field_name}")
        if field_name in ["3C286", "1328+307", "1331+305", "J1331+3030"]:
            ref_freq, I, polyI, poly_pfrac, poly_pangle = get_polmodel_coeff(
                datadir + "/3C286_pol_model.txt"
            )
        elif field_name in ["3C138", "0518+165", "0521+166", "J0521+1638"]:
            ref_freq, I, polyI, poly_pfrac, poly_pangle = get_polmodel_coeff(
                datadir + "/3C138_pol_model.txt"
            )
        else:
            print("Field name is not either of the polcal, 3C286 or 3C138.")
            return 1
        print(f"Reference frequency: {ref_freq} GHz.")
        with suppress_casa_output():
            setjy(
                vis=msname,
                field=field_name,
                scalebychan=True,
                standard="manual",
                fluxdensity=[I, 0.0, 0.0, 0.0],
                spix=polyI,
                reffreq=f"{ref_freq}GHz",
                polindex=poly_pfrac,
                polangle=poly_pangle,
                rotmeas=0,
                usescratch=True,
            )
    except Exception as e:
        traceback.print_exc()
    return


def phasecal_setjy(msname="", field="", ismms=False, n_threads=-1, dry_run=False):
    """
    Setjy phasecal fields

    Parameters
    ----------
    msname : str
        Measurement set
    field : str, optional
        Phasecal fields
    ismms : bool, optional
        Is a multi-ms ot not
    n_threads : int, optional
        Number of OpenMP threads
    """
    limit_threads(n_threads=n_threads)
    from casatasks import setjy

    if dry_run:
        process = psutil.Process(os.getpid())
        mem = round(process.memory_info().rss / 1024**3, 2)  # in GB
        return mem
    try:
        with suppress_casa_output():
            setjy(
                vis=msname,
                field=field,
                standard="manual",
                fluxdensity=[1.0, 0, 0, 0],
                usescratch=True,
                ismms=ismms,
            )
    except Exception as e:
        traceback.print_exc()
    return


def import_fluxcal_models(mslist, fluxcal_fields, fluxcal_scans, ncpus=1, mem_frac=0.8):
    """
    Import model visibilities using crystalball

    Parameters
    ----------
    mslist : list
        List of sub-MS
    fluxcal_fields : list
        Fluxcal fields
    fluxcal_scans : dict
        Fluxcal scans dict
    ncpus : int
        Number of CPU threads to use
    mem_frac : float
        Memory fraction to use

    Returns
    -------
    int
        Success message
    """
    try:
        if len(fluxcal_fields) == 0:
            print("No flux calibrator scan is present.\n")
            return 1
        print("##############################")
        print("Import fluxcal models")
        print("##############################")
        bandname = get_band_name(mslist[0])
        cpu_count = psutil.cpu_count()
        ncpus = max(1, ncpus)
        ncpus = min(ncpus, cpu_count - 1)
        mem_frac = max(mem_frac, 0.8)
        scans = []
        for ms in mslist:
            ms_scans = get_ms_scans(ms)
            for s in ms_scans:
                scans.append(s)
        for fluxcal in fluxcal_fields:
            f_scan = fluxcal_scans[fluxcal]
            for s in f_scan:
                if s in scans:
                    pos = scans.index(s)
                    sub_msname = mslist[pos]
                    modelname = datadir + "/" + fluxcal + "_" + bandname + "_model.txt"
                    crys_cmd_args = [
                        "-sm " + modelname,
                        "-f " + fluxcal,
                        "-ns 1000",
                        "-j " + str(ncpus),
                        "-mf " + str(mem_frac),
                    ]
                    crys_cmd = (
                        "crystalball " + " ".join(crys_cmd_args) + " " + sub_msname
                    )
                    print(crys_cmd)
                    tmpfile = f"tmp_{os.path.basename(sub_msname).split('.ms')[0]}"
                    with suppress_casa_output():
                        msg = os.system(crys_cmd + f" > {tmpfile}")
                    os.system(f"rm -rf {tmpfile}")
                    if msg == 0:
                        print(f"Fluxcal model is imported successfully for scan: {s}.")
                    else:
                        print(f"Error in importing fluxcal model for scan: {s}.")
        return 0
    except Exception as e:
        print("Error in importing model.")
        traceback.print_exc()
        return 1


def import_phasecal_models(
    mslist,
    phasecal_fields,
    phasecal_scans,
    workdir,
    cpu_frac=0.8,
    mem_frac=0.8,
    dask_addr=None,
):
    """
    Import model visibilities for phasecal

    Parameters
    ----------
    mslist : list
        List of sub-MS
    phasecal_fields : list
        Phasecal fields
    phasecal_scans : list
        Phasecal scans
    workdir : str
        Work directory
    cpu_frac : float, optional
        CPU fraction to use
    mem_frac : float, optional
        Memory fraction to use
    dask_addr : str, optional
        Dask scheduler address

    Returns
    -------
    int
        Success message
    """
    try:
        print("##########################")
        print("Import phasecal models")
        print("##########################")
        scans = []
        for ms in mslist:
            ms_scans = get_ms_scans(ms)
            for s in ms_scans:
                scans.append(s)
        mem_limit = phasecal_setjy(dry_run=True)
        if dask_addr is None:
            dask_client, dask_cluster, n_jobs, n_threads, mem_limit, dask_dir = (
                get_dask_client(
                    len(mslist),
                    dask_dir=workdir,
                    cpu_frac=cpu_frac,
                    mem_frac=mem_frac,
                    min_mem_per_job=mem_limit,
                )
            )
        else:
            _, _, n_jobs, n_threads, mem_limit, dask_dir = get_dask_client(
                len(mslist),
                dask_dir=workdir,
                cpu_frac=cpu_frac,
                mem_frac=mem_frac,
                min_mem_per_job=mem_limit,
                only_cal=True,
            )
            os.system(f"rm -rf {dask_dir}")
            dask_client = Client(address=dask_addr)
        wait_for_dask_workers(dask_client,min_worker=1,timeout=60)
        tasks = []
        for phasecal in phasecal_fields:
            ph_scan = phasecal_scans[phasecal]
            for s in ph_scan:
                if s in scans:
                    pos = scans.index(s)
                    sub_msname = mslist[pos]
                    tasks.append(
                        delayed(phasecal_setjy)(
                            sub_msname,
                            field=phasecal,
                            ismms=True,
                            n_threads=n_threads,
                        )
                    )
        futures = dask_client.compute(tasks)
        results = list(dask_client.gather(futures))
        dask_client.close()
        if dask_addr is None:
            dask_cluster.close()
            os.system(f"rm -rf {dask_dir}")
        print("Phasecal models are initiated.\n")
        return 0
    except Exception as e:
        print("Error in initiaing phasecal models.")
        traceback.print_exc()
        return 1


def import_polcal_models(
    mslist,
    polcal_fields,
    polcal_scans,
    workdir,
    cpu_frac=0.8,
    mem_frac=0.8,
    dask_addr=None,
):
    """
    Import model for polarization calibrators (3C286 or 3C138)

    Parameters
    ----------
    mslist : list
        List of sub-MS
    polcal_fields : list
        Polcal fields
    polcal_scans : list
        Polcal scans
    workdir : str
        Work directory
    n_threads : int, optional
        Number of OpenMP threads
    dask_addr : str, optional
        Dask scheduler address

    Returns
    -------
    int
        Success message
    """
    try:
        if len(polcal_scans) == 0:
            print("No polarization calibrator scans is present.")
            return 1
        scans = []
        for ms in mslist:
            ms_scans = get_ms_scans(ms)
            for s in ms_scans:
                scans.append(s)
        mem_limit = polcal_setjy(dry_run=True)
        if dask_addr is None:
            dask_client, dask_cluster, n_jobs, n_threads, mem_limit, dask_dir = (
                get_dask_client(
                    len(polcal_scans),
                    dask_dir=workdir,
                    cpu_frac=cpu_frac,
                    mem_frac=mem_frac,
                    min_mem_per_job=mem_limit,
                )
            )
        else:
            _, _, n_jobs, n_threads, mem_limit, dask_dir = get_dask_client(
                len(polcal_scans),
                dask_dir=workdir,
                cpu_frac=cpu_frac,
                mem_frac=mem_frac,
                min_mem_per_job=mem_limit,
                only_cal=True,
            )
            os.system(f"rm -rf {dask_dir}")
            dask_client = Client(address=dask_addr)
        wait_for_dask_workers(dask_client,min_worker=1,timeout=60)
        tasks = []
        for polcal_field in polcal_fields:
            p_scan = polcal_scans[polcal_field]
            for s in p_scan:
                if s in scans:
                    pos = scans.index(s)
                    sub_msname = mslist[pos]
                    tasks.append(
                        delayed(polcal_setjy)(
                            sub_msname, polcal_field, ismms=True, n_threads=n_threads
                        )
                    )
        futures = dask_client.compute(tasks)
        results = list(dask_client.gather(futures))
        dask_client.close()
        if dask_addr is None:
            dask_cluster.close()
            os.system(f"rm -rf {dask_dir}")
        return 0
    except Exception as e:
        traceback.print_exc()
        return 1


def import_all_models(msname, workdir, cpu_frac=0.8, mem_frac=0.8, dask_addr=None):
    """
    Import all calibrator models

    Parameters
    ----------
    msname : str
        Measurement set name
    workdir : str
        Work directory
    cpu_frac : float, optional
        CPU fraction to use
    mem_frac : float, optional
        Memory fraction to use
    dask_addr : str, optional
        Dask scheduler address

    Returns
    -------
    int
        Success message
    """
    start_time = time.time()
    cpu_threads = psutil.cpu_count()
    ncpu = int(cpu_threads * cpu_frac)
    mem_frac = 1 - mem_frac
    try:
        msname = msname.rstrip("/")
        correct_missing_col_subms(msname)
        mspath = os.path.dirname(os.path.abspath(msname))
        os.chdir(mspath)
        result = get_submsname_scans(msname)
        if result is not None:  # If multi-ms
            mslist, scans = result
        else:
            print("Please provide a multi-MS with scans partitioned.")
            return 1, 1, 1
        fluxcal_fields, fluxcal_scans = get_fluxcals(msname)
        phasecal_fields, phasecal_scans, phasecal_flux_list = get_phasecals(msname)
        polcal_fields, polcal_scans = get_polcals(msname)
        fluxcal_result = import_fluxcal_models(
            mslist, fluxcal_fields, fluxcal_scans, ncpus=ncpu, mem_frac=mem_frac
        )
        if fluxcal_result != 0:
            print("##################")
            print("Total time taken : " + str(time.time() - start_time) + "s")
            print("##################\n")
            return fluxcal_result, 1, 1
        phasecal_result = import_phasecal_models(
            mslist,
            phasecal_fields,
            phasecal_scans,
            workdir,
            cpu_frac=cpu_frac,
            mem_frac=mem_frac,
            dask_addr=dask_addr,
        )
        polcal_result = import_polcal_models(
            mslist,
            polcal_fields,
            polcal_scans,
            workdir,
            cpu_frac=cpu_frac,
            mem_frac=mem_frac,
            dask_addr=dask_addr,
        )
        if phasecal_result != 0:
            print(
                "Phasecal model was not import, but fluxcal model import is successful."
            )
        if polcal_result != 0:
            print(
                "Polcal model was not imported, but fluxcal model import is successful."
            )
        print("##################")
        print("Total time taken : " + str(time.time() - start_time) + "s")
        print("##################\n")
        return fluxcal_result, phasecal_result, polcal_result
    except Exception as e:
        print("##################")
        print("Total time taken : " + str(time.time() - start_time) + "s")
        print("##################\n")
        traceback.print_exc()
        return 1, 1, 1


def main(
    msname,
    workdir="",
    cpu_frac=0.8,
    mem_frac=0.8,
    logfile=None,
    jobid=0,
    start_remote_log=False,
    dask_addr=None,
):
    """
    Run the calibrator model import pipeline.

    Parameters
    ----------
    msname : str
        Path to the measurement set (MS) to process.
    workdir : str, optional
        Working directory for intermediate and output files. If empty, defaults
        to `<msname>/workdir`. Default is "".
    cpu_frac : float, optional
        Fraction of total CPU cores to use. Default is 0.8.
    mem_frac : float, optional
        Fraction of total system memory to use. Default is 0.8.
    logfile : str or None, optional
        Path to the logfile for logging messages. If None, logging to file is disabled.
    jobid : int, optional
        Job identifier for tracking and PID management. Default is 0.
    start_remote_log : bool, optional
        If True, enables remote logging based on credentials stored in the
        working directory. Default is False.
    dask_addr : str, optional
        Dask scheduler address

    Returns
    -------
    int
        Success message
    """
    pid = os.getpid()
    cachedir = get_cachedir()
    save_pid(pid, f"{cachedir}/pids/pids_{jobid}.txt")

    if workdir == "":
        workdir = os.path.dirname(os.path.abspath(msname)) + "/workdir"
    os.makedirs(workdir, exist_ok=True)

    ############
    # Logger
    ############
    observer = None
    if (
        start_remote_log
        and os.path.exists(f"{workdir}/jobname_password.npy")
        and logfile is not None
    ):
        time.sleep(5)
        jobname, password = np.load(
            f"{workdir}/jobname_password.npy", allow_pickle=True
        )
        if os.path.exists(logfile):
            observer = init_logger(
                "import_model", logfile, jobname=jobname, password=password
            )
    ###########

    try:
        if msname and os.path.exists(msname):
            fluxcal_result, phasecal_result, polcal_result = import_all_models(
                os.path.abspath(msname),
                workdir,
                cpu_frac=cpu_frac,
                mem_frac=mem_frac,
                dask_addr=dask_addr,
            )
            os.system("touch " + workdir + "/.fluxcal_" + str(fluxcal_result))
            os.system("touch " + workdir + "/.phasecal_" + str(phasecal_result))
            os.system("touch " + workdir + "/.polcal_" + str(polcal_result))
            if fluxcal_result == 1 or (
                fluxcal_result == 1 and phasecal_result == 1 and polcal_result == 1
            ):
                msg = 1
            else:
                msg = 0
        else:
            print("Please provide correct measurement set.\n")
            msg = 1
    except Exception as e:
        traceback.print_exc()
        os.system("touch " + workdir + "/.fluxcal_1")
        os.system("touch " + workdir + "/.phasecal_1")
        os.system("touch " + workdir + "/.polcal_1")
        msg = 1
    finally:
        time.sleep(5)
        drop_cache(msname)
        drop_cache(workdir)
        clean_shutdown(observer)
    return msg


def cli():
    usage = "Import calibrator models"
    parser = argparse.ArgumentParser(
        description=usage, formatter_class=SmartDefaultsHelpFormatter
    )

    # Essential parameters
    basic_args = parser.add_argument_group(
        "###################\nEssential parameters\n###################"
    )
    basic_args.add_argument("msname", type=str, help="Name of measurement set")
    basic_args.add_argument(
        "--workdir", type=str, default="", help="Name of work directory"
    )

    # Advanced parameters
    adv_args = parser.add_argument_group(
        "###################\nAdvanced parameters\n###################"
    )
    adv_args.add_argument(
        "--start_remote_log", action="store_true", help="Start remote logging"
    )

    # Resource management parameters
    hard_args = parser.add_argument_group(
        "###################\nHardware resource management parameters\n###################"
    )
    hard_args.add_argument(
        "--cpu_frac", type=float, default=0.8, help="CPU fraction to use"
    )
    hard_args.add_argument(
        "--mem_frac", type=float, default=0.8, help="Memory fraction to use"
    )
    hard_args.add_argument("--logfile", type=str, default=None, help="Log file")
    hard_args.add_argument("--jobid", type=int, default=0, help="Job ID")

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    args = parser.parse_args()

    msg = main(
        msname=args.msname,
        workdir=args.workdir,
        cpu_frac=args.cpu_frac,
        mem_frac=args.mem_frac,
        logfile=args.logfile,
        jobid=args.jobid,
        start_remote_log=args.start_remote_log,
    )
    return msg


if __name__ == "__main__":
    result = cli()
    print(
        "\n###################\nVisibility simulation is finished.\n###################\n"
    )
    os._exit(result)



import logging
import dask
import numpy as np
import argparse
import traceback
import warnings
import time
import glob
import sys
import os
from casatasks import casalog
from casatools import msmetadata, ms as casamstool
from dask import delayed
from dask.distributed import Client
from meersolar.utils import *

logging.getLogger("distributed").setLevel(logging.WARNING)

try:
    casalogfile = casalog.logfile()
    os.system("rm -rf " + casalogfile)
except BaseException:
    pass

datadir = get_datadir()


def make_solar_DS(
    msname,
    workdir,
    ds_file_name="",
    extension="png",
    target_scans=[],
    scans=[],
    merge_scan=False,
    showgui=False,
    cpu_frac=0.8,
    mem_frac=0.8,
    dask_addr=None,
):
    """
    Make solar dynamic spectrum and plots

    Parameters
    ----------
    msname : str
        Measurement set name'
    workdir : str
        Work directory
    ds_file_name : str, optional
        DS file name prefix
    extension : str, optional
        Image file extension
    target_scans : list, optional
        Target scans
    scans : list, optional
        Scan list
    merge_scan : bool, optional
        Merge scans in one plot or not
    showgui : bool, optional
        Show GUI
    cpu_frac : float, optional
        CPU fraction to use
    mem_frac : float, optional
        Memory fraction to use
    dask_addr : str, optional
        Dask scheduler address
    """
    warnings.filterwarnings("ignore", category=RuntimeWarning)
    os.makedirs(f"{workdir}/dynamic_spectra", exist_ok=True)
    print("##############################################")
    print(f"Start making dynamic spectra for ms: {msname}")
    print("##############################################")
    if len(target_scans) > 0:
        temp_target_scans = []
        for s in target_scans:
            temp_target_scans.append(int(s))
        target_scans = temp_target_scans

    ##################################
    # Making and ploting
    ##################################
    if len(scans) == 0:
        scans, cal_scans, f_scans, g_scans, p_scans = get_cal_target_scans(msname)
    valid_scans = get_valid_scans(msname)
    final_scans = []
    scan_size_list = []
    msmd = msmetadata()
    mstool = casamstool()
    for scan in scans:
        if scan in valid_scans:
            if len(target_scans) == 0 or (
                len(target_scans) > 0 and int(scan) in target_scans
            ):
                final_scans.append(int(scan))
                msmd.open(msname)
                nchan = msmd.nchan(0)
                nant = msmd.nantennas()
                msmd.close()
                mstool.open(msname)
                mstool.select({"scan_number": int(scan)})
                nrow = mstool.nrow(True)
                mstool.close()
                nbaselines = int(nant + (nant * (nant - 1) / 2))
                scan_size = (5 * (nrow / nbaselines) * 16) / (1024**3)
                scan_size_list.append(scan_size)
    if len(final_scans) == 0:
        print("No scans to make dynamic spectra.")
        return
    del scans
    scans = sorted(final_scans)
    print(f"Scans: {scans}")
    msname = msname.rstrip("/")
    if ds_file_name == "":
        ds_file_name = os.path.basename(msname).split(".ms")[0] + "_DS"
    hascor = check_datacolumn_valid(msname, datacolumn="CORRECTED_DATA")
    if hascor:
        datacolumn = "CORRECTED_DATA"
    else:
        datacolumn = "DATA"
    mspath = os.path.dirname(msname)
    mem_limit = max(scan_size_list)
    if dask_addr is None:
        dask_client, dask_cluster, n_jobs, n_threads, mem_limit, dask_dir = get_dask_client(
            len(scans),
            dask_dir=workdir,
            cpu_frac=cpu_frac,
            mem_frac=mem_frac,
            min_mem_per_job=mem_limit,
        )
    else:
        _, _, n_jobs, n_threads, mem_limit, dask_dir = get_dask_client(
            len(scans),
            dask_dir=workdir,
            cpu_frac=cpu_frac,
            mem_frac=mem_frac,
            min_mem_per_job=mem_limit,
            only_cal=True,
        )
        dask_client=Client(address=dask_addr)
        os.system(f"rm -rf {dask_dir}")
    wait_for_dask_workers(dask_client,min_worker=1,timeout=60)
    tasks = []
    for scan in scans:
        tasks.append(
            delayed(make_ds_file_per_scan)(
                msname,
                f"{workdir}/dynamic_spectra/{ds_file_name}_scan_{scan}",
                scan,
                datacolumn,
            )
        )
    futures = dask_client.compute(tasks)
    results = list(dask_client.gather(futures))
    dask_client.close()
    if dask_addr is None:
        dask_cluster.close()
        os.system(f"rm -rf {dask_dir}")
    ds_files = [
        f"{workdir}/dynamic_spectra/{ds_file_name}_scan_{scan}.npy" for scan in scans
    ]
    print(f"DS files: {ds_files}")
    if not merge_scan:
        plots = []
        for dsfile in ds_files:
            plot_file = make_ds_plot(
                [dsfile],
                plot_file=dsfile.replace(".npy", f".{extension}"),
                showgui=showgui,
            )
            plots.append(plot_file)
    else:
        plot_file = make_ds_plot(
            ds_files,
            plot_file=f"{workdir}/dynamic_spectra/{ds_file_name}.{extension}",
            showgui=showgui,
        )
    goes_files = glob.glob(f"{workdir}/dynamic_spectra/sci*.nc")
    for f in goes_files:
        os.system(f"rm -rf {f}")
    return


def make_dsfiles(
    msname,
    workdir,
    outdir,
    extension="png",
    target_scans=[],
    merge_scans=True,
    seperate_scans=True,
    cpu_frac=0.8,
    mem_frac=0.8,
    dask_addr=None,
):
    """
    Make all dynamic spectra of the solar scans

    Parameters
    ----------
    msname : str
        Measurement set name
    workdir : str
        Work directory
    outdir : str
        Output directory
    extension : str, optional
        Plot file extension
    target_scans : list, optional
        Target scans
    merge_scans: bool, optional
        Merge scans
    seperate_scans : bool, optional
        Seperate scans
    cpu_frac : float, optional
        CPU fraction to use
    mem_frac : float, optional
        Memory fraction to use
    dask_addr : str, optional
        Dask scheduler address

    Returns
    -------
    list
        Dynamic spectra files
    """
    msname = msname.rstrip("/")
    workdir = workdir.rstrip("/")
    if seperate_scans == False and merge_scans == False:
        return
    try:
        if seperate_scans:
            make_solar_DS(
                msname,
                workdir,
                extension=extension,
                target_scans=target_scans,
                merge_scan=False,
                cpu_frac=cpu_frac,
                mem_frac=mem_frac,
                dask_addr=dask_addr,
            )
        if merge_scans:
            make_solar_DS(
                msname,
                workdir,
                extension=extension,
                target_scans=target_scans,
                merge_scan=True,
                cpu_frac=cpu_frac,
                mem_frac=mem_frac,
                dask_addr=dask_addr,
            )
        if os.path.samefile(outdir, workdir) == False:
            os.makedirs(f"{outdir}/dynamic_spectra",exist_ok=True)
            os.system(f"mv {workdir}/dynamic_spectra/* {outdir}/dynamic_spectra/")
            os.system(f"rm -rf {workdir}/dynamic_spectra")
        ds_file_name = os.path.basename(msname).split(".ms")[0] + "_DS"
        ds_files = glob.glob(f"{outdir}/dynamic_spectra/{ds_file_name}*.{extension}")
        return ds_files
    except Exception as e:
        traceback.print_exc()
        return []
    finally:
        time.sleep(5)
        drop_cache(msname)
        drop_cache(workdir)


def main(
    msname,
    workdir,
    outdir,
    extension="png",
    target_scans=[],
    merge=True,
    seperate=True,
    cpu_frac=0.8,
    mem_frac=0.8,
    logfile=None,
    jobid="0",
    start_remote_log=False,
    dask_addr=None,
):
    """
    Make dynamic spectra

    Parameters
    ----------
    msname : str
        Measurement set
    workdir : str
        Work directory
    outdir : str
        Output directory
    extension : str, optional
        Plot extension
    target_scans : list, optional
        Target scans
    merge : bool, optional
        Merge scans plot
    seperate : bool, optional
        Seperate scan plots
    cpu_frac : float, optional
        CPU fraction
    mem_frac : float, optional
        Memory fraction
    logfile : str, optional
        Log file
    jobid : str, optional
        Job ID
    start_remote_log : bool, optional
        Start remote log
    dask_addr: str, optional
        Dask scheduler address

    Returns
    -------
    int
        Success messsage
    """
    pid = os.getpid()
    cachedir = get_cachedir()
    save_pid(pid, f"{cachedir}/pids/pids_{jobid}.txt")

    if workdir == "":
        workdir = os.path.dirname(os.path.abspath(msname)) + "/workdir"
    os.makedirs(workdir, exist_ok=True)

    ############
    # Logger
    ############
    observer = None
    if (
        start_remote_log
        and os.path.exists(f"{workdir}/jobname_password.npy")
        and logfile is not None
    ):
        time.sleep(5)
        jobname, password = np.load(
            f"{workdir}/jobname_password.npy", allow_pickle=True
        )
        if os.path.exists(logfile):
            observer = init_logger(
                "ds_plot", logfile, jobname=jobname, password=password
            )
    ###########

    try:
        if msname != "" and os.path.exists(msname):
            ds_files = make_dsfiles(
                msname,
                workdir,
                outdir,
                extension=extension,
                target_scans=target_scans,
                merge_scans=merge,
                seperate_scans=seperate,
                cpu_frac=float(cpu_frac),
                mem_frac=float(mem_frac),
                dask_addr=dask_addr,
            )
            msg = 0
        else:
            print("Please provide a valid measurement set.\n")
            msg = 1
    except Exception as e:
        traceback.print_exc()
        msg = 1
    finally:
        time.sleep(5)
        clean_shutdown(observer)
    return msg


def cli():
    parser = argparse.ArgumentParser(
        description="Make dynamic spectra of solar scans",
        formatter_class=SmartDefaultsHelpFormatter,
    )
    # === Essential parameters ===
    essential = parser.add_argument_group(
        "###################\nEssential parameters\n###################"
    )
    essential.add_argument("msname", type=str, help="Measurement set name")
    essential.add_argument(
        "--workdir",
        type=str,
        dest="workdir",
        required=True,
        help="Working directory",
    )
    essential.add_argument(
        "--outdir",
        type=str,
        dest="outdir",
        required=True,
        help="Output directory",
    )

    # === Advanced parameters ===
    adv_args = parser.add_argument_group(
        "###################\nAdvanced parameters\n###################"
    )
    adv_args.add_argument(
        "--extension",
        type=str,
        default="png",
        help="Save file extension",
    )
    adv_args.add_argument(
        "--target_scans",
        nargs="*",
        type=str,
        default=[],
        help="List of target scans to process (space-separated, e.g. 3 5 7)",
    )
    adv_args.add_argument(
        "--no_merge",
        action="store_false",
        dest="merge",
        help="Do not merge scans",
    )
    adv_args.add_argument(
        "--no_seperate",
        action="store_false",
        dest="seperate",
        help="Do not seperate scans",
    )
    adv_args.add_argument(
        "--start_remote_log", action="store_true", help="Start remote logging"
    )

    # === Advanced local system/ per node hardware resource parameters ===
    hard_args = parser.add_argument_group(
        "###################\nHardware resource management parameters\n###################"
    )
    hard_args.add_argument(
        "--cpu_frac",
        type=float,
        default=0.8,
        help="Fraction of CPU usuage per node",
    )
    hard_args.add_argument(
        "--mem_frac",
        type=float,
        default=0.8,
        help="Fraction of memory usuage per node",
    )
    hard_args.add_argument(
        "--logfile", type=str, default=None, help="Optional path to log file"
    )
    hard_args.add_argument(
        "--jobid", type=str, default="0", help="Job ID for logging and PID tracking"
    )

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    args = parser.parse_args()

    msg = main(
        msname=args.msname,
        workdir=args.workdir,
        outdir=args.outdir,
        extension=args.extension,
        target_scans=args.target_scans,
        merge=args.merge,
        seperate=args.seperate,
        cpu_frac=args.cpu_frac,
        mem_frac=args.mem_frac,
        logfile=args.logfile,
        jobid=args.jobid,
        start_remote_log=args.start_remote_log,
    )
    return msg


if __name__ == "__main__":
    result = cli()
    print(
        "\n###################\nDynamic spectra are produced successfully.\n###################\n"
    )
    os._exit(result)



# -*- coding: utf-8 -*-

import logging
import sys

from loguru import logger


class InterceptHandler(logging.Handler):
    """Intercept logging messages and reroute them to the loguru."""

    def emit(self, record):
        # Get corresponding Loguru level if it exists
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno

        # Find caller from where originated the logged message
        frame, depth = logging.currentframe(), 2
        while frame.f_code.co_filename == logging.__file__:
            frame = frame.f_back
            depth += 1

        (
            logger.opt(depth=depth, exception=record.exc_info).log(
                level, record.getMessage()
            )
        )


logging.basicConfig(handlers=[InterceptHandler()], level=0)


# Put together a formatting string for the logger.
# Split into pieces to improve legibility.
tim_fmt = "<green>{time:YYYY-MM-DD HH:mm:ss}</green>"
lvl_fmt = "<level>{level: <8}</level>"
src_fmt = "<cyan>{module}</cyan>:<cyan>{function}</cyan>"
msg_fmt = "<level>{message}</level>"

fmt = " | ".join([tim_fmt, lvl_fmt, src_fmt, msg_fmt])

config = {
    "handlers": [
        {"sink": sys.stderr, "level": "INFO", "format": fmt},
        # {"sink": "{time:YYYYMMDD_HHmmss}_crystalball.log",
        #  "level": "DEBUG",
        #  "format": fmt,
        #  }
    ],
}

logger.configure(**config)



import logging
import psutil
import dask
import numpy as np
import argparse
import traceback
import time
import glob
import sys
import os
from casatasks import casalog
from casatools import msmetadata
from dask import delayed
from dask.distributed import Client
from meersolar.utils import *

logging.getLogger("distributed").setLevel(logging.WARNING)

try:
    casalogfile = casalog.logfile()
    os.system("rm -rf " + casalogfile)
except BaseException:
    pass

datadir = get_datadir()


def single_ms_flag(
    msname="",
    badspw="",
    bad_ants_str="",
    datacolumn="data",
    use_tfcrop=True,
    use_rflag=False,
    flagdimension="freqtime",
    flag_autocorr=True,
    n_threads=-1,
    memory_limit=-1,
    dry_run=False,
):
    """
    Flag on a single ms

    Parameters
    ----------
    msname : str
        Measurement set name
    badspw : str, optional
        Bad spectral window
    bad_ants_str : str, optional
        Bad antenna string
    datacolumn : str, optional
        Data column
    use_tfcrop : str, optional
        Use tfcrop or not
    use_rflag : str, optional
        Use rflag or not
    flagdimension : str, optional
        Flag dimension (only applicable for tfcrop)
    flag_autocorr : bool, optional
        Flag autocorrelations or not
    n_threads : int, optional
        Number of OpenMP threads
    memory_limit : float, optional
        Memory limit in GB
    dry_run : bool, optional
        Return the amount of pre-occupied memory in GB
    """
    limit_threads(n_threads=n_threads)
    from casatasks import flagdata

    if dry_run:
        process = psutil.Process(os.getpid())
        mem = round(process.memory_info().rss / 1024**3, 2)  # in GB
        return mem
    msname = msname.rstrip("/")
    try:
        ##############################
        # Flagging bad channels
        ##############################
        if badspw != "":
            print(f"Flagging bad spectral windows: {badspw}\n")
            try:
                with suppress_casa_output():
                    flagdata(
                        vis=msname,
                        mode="manual",
                        spw=badspw,
                        cmdreason="badchan",
                        flagbackup=False,
                    )
            except BaseException:
                pass

        ##############################
        # Flagging bad antennas
        ##############################
        if bad_ants_str != "":
            print(f"Flagging bad antenna: {bad_ants_str}\n")
            try:
                with suppress_casa_output():
                    flagdata(
                        vis=msname,
                        mode="manual",
                        antenna=bad_ants_str,
                        cmdreason="badant",
                        flagbackup=False,
                    )
            except BaseException:
                pass

        #################################
        # Clip zero amplitude data points
        #################################
        try:
            with suppress_casa_output():
                flagdata(
                    vis=msname,
                    mode="clip",
                    clipzeros=True,
                    datacolumn=datacolumn,
                    autocorr=flag_autocorr,
                    flagbackup=False,
                )
        except BaseException:
            pass

        #################################
        # Flag auto-correlations
        #################################
        if flag_autocorr:
            try:
                with suppress_casa_output():
                    flagdata(
                        vis=msname,
                        mode="manual",
                        autocorr=True,
                        datacolumn=datacolumn,
                        flagbackup=False,
                    )
            except BaseException:
                pass

        ####################################################
        # Check if required columns are present for residual
        ####################################################
        if datacolumn == "residual" or datacolumn == "RESIDUAL":
            modelcolumn_present = check_datacolumn_valid(
                msname, datacolumn="MODEL_DATA"
            )
            corcolumn_present = check_datacolumn_valid(
                msname, datacolumn="CORRECTED_DATA"
            )
            if modelcolumn_present == False or corcolumn_present == False:
                datacolumn = "corrected"
        elif datacolumn == "RESIDUAL_DATA":
            modelcolumn_present = check_datacolumn_valid(
                msname, datacolumn="MODEL_DATA"
            )
            datacolumn_present = check_datacolumn_valid(msname, datacolumn="DATA")
            if modelcolumn_present == False or datacolumn_present == False:
                datacolumn = "corrected"

        #################################################
        # Whether corrected data column is present or not
        #################################################
        if datacolumn == "corrected" or datacolumn == "CORRECTED_DATA":
            corcolumn_present = check_datacolumn_valid(
                msname, datacolumn="CORRECTED_DATA"
            )
            if not corcolumn_present:
                print(
                    "Corrected data column is chosen for flagging, but it is not present.\n"
                )
                return
            else:
                datacolumn = "corrected"

        #################################################
        # Whether data column is present or not
        #################################################
        if datacolumn == "data" or datacolumn == "DATA":
            datacolumn_present = check_datacolumn_valid(msname, datacolumn="DATA")
            if not datacolumn_present:
                print("Data column is chosen for flagging, but it is not present.\n")
                return
            else:
                datacolumn = "data"

        ###########################
        # Determinign time chunking
        ############################
        if use_tfcrop or use_rflag:
            nchunk = get_chunk_size(msname, memory_limit=memory_limit)
            if nchunk <= 1:
                ntime = "scan"
            else:
                msmd = msmetadata()
                msmd.open(msname)
                scan = np.unique(msmd.scannumbers())[0]
                times = msmd.timesforspws(0)
                msmd.close()
                total_time = np.nanmax(times) - np.nanmin(times)
                timeres = np.nanmin(np.diff(times))
                ntime = float(total_time / nchunk)
                if ntime < timeres:
                    ntime = timeres

        ##############
        # Tfcrop flag
        ##############
        if use_tfcrop:
            try:
                with suppress_casa_output():
                    flagdata(
                        vis=msname,
                        mode="tfcrop",
                        timefit="line",
                        freqfit="line",
                        extendflags=False,
                        flagdimension=flagdimension,
                        timecutoff=5.0,
                        freqcutoff=5.0,
                        extendpols=True,
                        growaround=False,
                        action="apply",
                        flagbackup=False,
                        overwrite=True,
                        writeflags=True,
                        datacolumn=datacolumn,
                        ntime=ntime,
                    )
            except BaseException:
                pass

        #############
        # Rflag flag
        #############
        try:
            with suppress_casa_output():
                flagdata(
                    vis=msname,
                    mode="rflag",
                    timefit="line",
                    freqfit="line",
                    extendflags=False,
                    timedevscale=5.0,
                    freqdevscale=5.0,
                    extendpols=True,
                    growaround=False,
                    action="apply",
                    flagbackup=False,
                    overwrite=True,
                    writeflags=True,
                    datacolumn=datacolumn,
                    ntime=ntime,
                )
        except BaseException:
            pass

        ##############
        # Extend flag
        ##############
        try:
            with suppress_casa_output():
                flagdata(
                    vis=msname,
                    mode="extend",
                    datacolumn="data",
                    clipzeros=True,
                    extendflags=False,
                    extendpols=True,
                    growtime=80.0,
                    growfreq=80.0,
                    growaround=False,
                    flagneartime=False,
                    flagnearfreq=False,
                    action="apply",
                    flagbackup=False,
                    overwrite=True,
                    writeflags=True,
                    ntime=ntime,
                )
        except BaseException:
            pass
        return 0
    except Exception as e:
        traceback.print_exc()
        return 1


def do_flagging(
    msname,
    workdir,
    datacolumn="data",
    flag_bad_ants=True,
    flag_bad_spw=True,
    use_tfcrop=True,
    use_rflag=False,
    flagdimension="freqtime",
    flag_autocorr=True,
    flag_backup=True,
    cpu_frac=0.8,
    mem_frac=0.8,
    dask_addr=None,
):
    """
    Function to perform initial flagging

    Parameters
    ----------
    msname : str
        Name of the ms
    workdir : str
        Work directory
    datacolumn : str, optional
        Data column
    flag_bad_ants : bool, optional
        Flag bad antennas
    flag_bad_spw : bool, optional
        Flag bad channels
    use_tfcrop : bool, optional
        Use tfcrop or not
    use_rflag : bool, optional
        Use rflag or not
    flagdimension : str, optional
        Flag dimension (only for tfcrop)
    flag_autocorr : bool,optional
        Flag auto-correlations
    flag_backup : bool, optional
        Flag backup
    cpu_frac : float, optional
        CPU fraction to use
    mem_frac : float, optional
        Memory fraction to use
    dask_addr : str, optional
        Dask scheduler address

    Returns
    -------
    int
        Success message
    """
    start_time = time.time()
    try:
        from casatasks import flagdata

        msname = msname.rstrip("/")
        mspath = os.path.dirname(os.path.abspath(msname))
        os.chdir(mspath)
        print("###########################")
        print("Flagging measurement set : ", msname)
        print("###########################\n")
        total_cpus = psutil.cpu_count(logical=True)
        ncpu = int(total_cpus * cpu_frac)
        correct_missing_col_subms(msname)
        print("Restoring all previous flags...")
        with suppress_casa_output():
            flagdata(vis=msname, mode="unflag", spw="0", flagbackup=False)
        fluxcal_field, fluxcal_scans = get_fluxcals(msname)
        if len(fluxcal_field) == 0:
            flag_bad_spw = False
            flag_bad_ants = False
        if flag_bad_spw:
            badspw = get_bad_chans(msname)
        else:
            bandspw = ""
        if flag_bad_ants:
            bad_ants, bad_ants_str = get_bad_ants(msname, fieldnames=fluxcal_field)
        else:
            bad_ants_str = ""
        ###########################
        # Dask local cluster setup
        ##########################
        if os.path.exists(msname + "/SUBMSS"):
            subms_list = glob.glob(msname + "/SUBMSS/*")
            for subms in subms_list:
                os.system(f"rm -rf {subms}/.flagversions")
        else:
            subms_list = [msname]
        ms_size_list = [get_column_size(ms) for ms in subms_list]
        mem_limit = max(ms_size_list)
        if dask_addr is None:
            dask_client, dask_cluster, n_jobs, n_threads, mem_limit, dask_dir = (
                get_dask_client(
                    len(subms_list),
                    dask_dir=workdir,
                    cpu_frac=cpu_frac,
                    mem_frac=mem_frac,
                    min_mem_per_job=mem_limit,
                )
            )
        else:
            _, _, n_jobs, n_threads, mem_limit, dask_dir = get_dask_client(
                len(subms_list),
                dask_dir=workdir,
                cpu_frac=cpu_frac,
                mem_frac=mem_frac,
                min_mem_per_job=mem_limit,
                only_cal=True,
            )
            os.system(f"rm -rf {dask_dir}")
            dask_client = Client(address=dask_addr)
        wait_for_dask_workers(dask_client,min_worker=1,timeout=60)
        if flag_backup:
            do_flag_backup(msname, flagtype="flagdata")
        tasks = [
            delayed(single_ms_flag)(
                ms,
                badspw=badspw,
                bad_ants_str=bad_ants_str,
                datacolumn=datacolumn,
                use_tfcrop=use_tfcrop,
                use_rflag=use_rflag,
                flagdimension=flagdimension,
                flag_autocorr=flag_autocorr,
                n_threads=n_threads,
                memory_limit=mem_limit,
            )
            for ms in subms_list
        ]
        futures = dask_client.compute(tasks)
        results = list(dask_client.gather(futures))
        dask_client.close()
        if dask_addr is None:
            dask_cluster.close()
            os.system(f"rm -rf {dask_dir}")
        print("##################")
        print("Total time taken : ", time.time() - start_time)
        print("##################\n")
        return 0
    except Exception as e:
        traceback.print_exc()
        print("##################")
        print("Total time taken : " + str(time.time() - start_time) + "s")
        print("##################\n")
        return 1


def main(
    msname,
    workdir="",
    datacolumn="DATA",
    flag_bad_ants=True,
    flag_bad_spw=True,
    use_tfcrop=False,
    use_rflag=False,
    flag_autocorr=True,
    flagbackup=True,
    flagdimension="freqtime",
    cpu_frac=0.8,
    mem_frac=0.8,
    logfile=None,
    jobid=0,
    start_remote_log=False,
    dask_addr=None,
):
    """
    Run the flagging pipeline for a measurement set.

    Parameters
    ----------
    msname : str
        Path to the input measurement set (MS) to be flagged.
    workdir : str, optional
        Working directory to store logs and temporary files. If empty, defaults to
        `<msname>/workdir`. Default is "".
    datacolumn : str, optional
        Data column to be flagged (e.g., "DATA", "CORRECTED"). Default is "DATA".
    flag_bad_ants : bool, optional
        If True, flags known bad antennas using pre-defined heuristics. Default is True.
    flag_bad_spw : bool, optional
        If True, flags bad spectral windows based on statistics. Default is True.
    use_tfcrop : bool, optional
        If True, applies the `tfcrop` automated flagging algorithm. Default is False.
    use_rflag : bool, optional
        If True, applies the `rflag` automated flagging algorithm. Default is False.
    flag_autocorr : bool, optional
        If True, flags auto-correlations. Default is True.
    flagbackup : bool, optional
        If True, saves a flag backup before applying new flags. Default is True.
    flagdimension : str, optional
        Dimension over which to apply automated flagging (e.g., "freqtime"). Default is "freqtime".
    cpu_frac : float, optional
        Fraction of total CPU resources to use. Default is 0.8.
    mem_frac : float, optional
        Fraction of total memory resources to use. Default is 0.8.
    logfile : str or None, optional
        Path to the log file for saving logs. If None, logging to file is skipped.
    jobid : int, optional
        Numeric job ID used for PID tracking. Default is 0.
    start_remote_log : bool, optional
        Whether to enable remote logging using credentials in the workdir. Default is False.
    dask_addr : str, optional
        Dask scheduler address

    Returns
    -------
    int
        Success message
    """
    pid = os.getpid()
    cachedir = get_cachedir()
    save_pid(pid, f"{cachedir}/pids/pids_{jobid}.txt")

    if workdir == "":
        workdir = os.path.dirname(os.path.abspath(msname)) + "/workdir"
    os.makedirs(workdir, exist_ok=True)

    ############
    # Logger
    ############
    observer = None
    if (
        start_remote_log
        and os.path.exists(f"{workdir}/jobname_password.npy")
        and logfile is not None
    ):
        time.sleep(5)
        jobname, password = np.load(
            f"{workdir}/jobname_password.npy", allow_pickle=True
        )
        if os.path.exists(logfile):
            observer = init_logger(
                "do_flagging", logfile, jobname=jobname, password=password
            )
    if observer == None:
        print("Remote link or jobname is blank. Not transmiting to remote logger.")
    ###########

    try:
        if msname and os.path.exists(msname):
            msg = do_flagging(
                msname,
                workdir,
                datacolumn=datacolumn,
                flag_bad_ants=flag_bad_ants,
                flag_bad_spw=flag_bad_spw,
                use_tfcrop=use_tfcrop,
                use_rflag=use_rflag,
                flagdimension=flagdimension,
                flag_autocorr=flag_autocorr,
                flag_backup=flagbackup,
                cpu_frac=cpu_frac,
                mem_frac=mem_frac,
                dask_addr=dask_addr,
            )
        else:
            print("Please provide correct measurement set.\n")
            msg = 1
    except Exception as e:
        traceback.print_exc()
        msg = 1
    finally:
        time.sleep(5)
        drop_cache(msname)
        drop_cache(workdir)
        clean_shutdown(observer)
    return msg


def cli():
    usage = "Initial flagging of calibrator data"
    parser = argparse.ArgumentParser(
        description=usage, formatter_class=SmartDefaultsHelpFormatter
    )

    # Essential parameters
    basic_args = parser.add_argument_group(
        "###################\nEssential parameters\n###################"
    )
    basic_args.add_argument("msname", type=str, help="Name of measurement set")
    basic_args.add_argument(
        "--workdir", type=str, default="", help="Name of work directory"
    )
    basic_args.add_argument(
        "--datacolumn", type=str, default="DATA", help="Name of the datacolumn"
    )

    # Advanced switches
    adv_args = parser.add_argument_group(
        "###################\nAdvanced parameters\n###################"
    )
    adv_args.add_argument(
        "--no_flag_bad_ants",
        dest="flag_bad_ants",
        action="store_false",
        help="Do not flag bad antennas",
    )
    adv_args.add_argument(
        "--no_flag_bad_spw",
        dest="flag_bad_spw",
        action="store_false",
        help="Do not flag bad spectral windows",
    )
    adv_args.add_argument(
        "--use_tfcrop", action="store_true", help="Use tfcrop flagging"
    )
    adv_args.add_argument("--use_rflag", action="store_true", help="Use rflag flagging")
    adv_args.add_argument(
        "--no_flag_autocorr",
        dest="flag_autocorr",
        action="store_false",
        help="Do not flag auto-correlations",
    )
    adv_args.add_argument(
        "--no_flagbackup",
        dest="flagbackup",
        action="store_false",
        help="Do not backup flags",
    )
    adv_args.add_argument(
        "--start_remote_log", action="store_true", help="Start remote logging"
    )

    # Resource management parameters
    hard_args = parser.add_argument_group(
        "###################\nHardware resource management parameters\n###################"
    )
    hard_args.add_argument(
        "--flagdimension", type=str, default="freqtime", help="Flag dimension"
    )
    hard_args.add_argument(
        "--cpu_frac", type=float, default=0.8, help="CPU fraction to use"
    )
    hard_args.add_argument(
        "--mem_frac", type=float, default=0.8, help="Memory fraction to use"
    )
    hard_args.add_argument("--logfile", type=str, default=None, help="Log file")
    hard_args.add_argument("--jobid", type=int, default=0, help="Job ID")

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    args = parser.parse_args()

    msg = main(
        msname=args.msname,
        workdir=args.workdir,
        datacolumn=args.datacolumn,
        flag_bad_ants=args.flag_bad_ants,
        flag_bad_spw=args.flag_bad_spw,
        use_tfcrop=args.use_tfcrop,
        use_rflag=args.use_rflag,
        flag_autocorr=args.flag_autocorr,
        flagbackup=args.flagbackup,
        flagdimension=args.flagdimension,
        cpu_frac=args.cpu_frac,
        mem_frac=args.mem_frac,
        logfile=args.logfile,
        jobid=args.jobid,
        start_remote_log=args.start_remote_log,
    )
    return msg


if __name__ == "__main__":
    result = cli()
    print(f"Final msg : {result}")
    print("\n###################\nBasic flagging is finished.\n###################\n")
    os._exit(result)



import types
import psutil
import numpy as np
import glob
import os
import traceback
import warnings
import astropy.units as u
from sunpy.net import Fido, attrs as a
from sunpy.timeseries import TimeSeries
from astroquery.jplhorizons import Horizons
from astropy.visualization import ImageNormalize, LogStretch
from astropy.wcs import FITSFixedWarning
from astropy.io import fits
from astropy.time import Time
from astropy.coordinates import SkyCoord
from casatools import msmetadata, ms as casamstool, table
from datetime import datetime as dt, timedelta
from .basic_utils import *
from .resource_utils import *
from .udocker_utils import *
from .ms_metadata import *
from .image_utils import *

warnings.simplefilter("ignore", category=FITSFixedWarning)


#######################
# MS metadata related
#######################


def get_fluxcals(msname):
    """
    Get fluxcal field names and scans (all scans, valids and invalids

    Parameters
    ----------
    msname : str
        Name of the ms

    Returns
    -------
    list
        Fluxcal field names
    dict
        Fluxcal scans
    """
    msmd = msmetadata()
    if os.path.exists(msname + "/SUBMSS"):
        mslist = glob.glob(msname + "/SUBMSS/*.ms")
    else:
        mslist = [msname]
    fluxcal_fields = []
    fluxcal_scans = {}
    for msname in mslist:
        msmd.open(msname)
        field_names = msmd.fieldnames()
        for field in field_names:
            if field in ["J1939-6342", "J0408-6545"]:
                if field not in fluxcal_fields:
                    fluxcal_fields.append(field)
                scans = msmd.scansforfield(field).tolist()
                if field in fluxcal_scans:
                    for scan in scans:
                        fluxcal_scans[field].append(scan)
                else:
                    fluxcal_scans[field] = scans
    msmd.close()
    msmd.done()
    del msmd
    for field in fluxcal_scans:
        scans = np.unique(fluxcal_scans[field]).tolist()
        fluxcal_scans[field] = scans
    return fluxcal_fields, fluxcal_scans


def get_polcals(msname):
    """
    Get polarization calibrator field names and scans (all scans, valids and invalids

    Parameters
    ----------
    msname : str
        Name of the ms

    Returns
    -------
    list
        Polcal field names
    dict
        Polcal scans
    """
    msmd = msmetadata()
    if os.path.exists(msname + "/SUBMSS"):
        mslist = glob.glob(msname + "/SUBMSS/*.ms")
    else:
        mslist = [msname]
    polcal_fields = []
    polcal_scans = {}
    for msname in mslist:
        msmd.open(msname)
        field_names = msmd.fieldnames()
        for field in field_names:
            if field in ["3C286", "1328+307", "1331+305", "J1331+3030"] or field in [
                "3C138",
                "0518+165",
                "0521+166",
                "J0521+1638",
            ]:
                if field not in polcal_fields:
                    polcal_fields.append(field)
                scans = msmd.scansforfield(field).tolist()
                if field in polcal_scans:
                    for scan in scans:
                        polcal_scans[field].append(scan)
                else:
                    polcal_scans[field] = scans
    msmd.close()
    msmd.done()
    del msmd
    for field in polcal_scans:
        scans = np.unique(polcal_scans[field]).tolist()
        polcal_scans[field] = scans
    return polcal_fields, polcal_scans


def get_phasecals(msname):
    """
    Get phasecal field names and scans (all scans, valids and invalids)

    Parameters
    ----------
    msname : str
        Name of the ms

    Returns
    -------
    list
        Phasecal field names
    dict
        Phasecal scans
    dict
        Phasecal flux
    """
    msmd = msmetadata()
    if os.path.exists(msname + "/SUBMSS"):
        mslist = glob.glob(msname + "/SUBMSS/*.ms")
    else:
        mslist = [msname]
    phasecal_fields = []
    phasecal_scans = {}
    phasecal_flux_list = {}
    datadir = get_datadir()
    for msname in mslist:
        msmd.open(msname)
        field_names = msmd.fieldnames()
        bandname = get_band_name(msname)
        if bandname == "U":
            phasecals, phasecal_flux = np.load(
                datadir + "/UHF_band_cal.npy", allow_pickle=True
            ).tolist()
        elif bandname == "L":
            phasecals, phasecal_flux = np.load(
                datadir + "/L_band_cal.npy", allow_pickle=True
            ).tolist()
        for field in field_names:
            if field in phasecals and (field != "J1939-6342" and field != "J0408-6545"):
                if field not in phasecal_fields:
                    phasecal_fields.append(field)
                scans = msmd.scansforfield(field).tolist()
                if field in phasecal_scans:
                    for scan in scans:
                        phasecal_scans[field].append(scan)
                else:
                    phasecal_scans[field] = scans
                flux = phasecal_flux[phasecals.index(field)]
                phasecal_flux_list[field] = flux
    msmd.close()
    msmd.done()
    del msmd
    for field in phasecal_scans:
        scans = np.unique(phasecal_scans[field]).tolist()
        phasecal_scans[field] = scans
    return phasecal_fields, phasecal_scans, phasecal_flux_list


def get_valid_scans(msname, field="", min_scan_time=1, n_threads=-1):
    """
    Get valid list of scans

    Parameters
    ----------
    msname : str
        Measurement set name
    field : str
        Field names (comma seperated)
    min_scan_time : float
        Minimum valid scan time in minute

    Returns
    -------
    list
        Valid scan list
    """
    limit_threads(n_threads=n_threads)
    from casatools import ms as casamstool

    mstool = casamstool()
    mstool.open(msname)
    scan_summary = mstool.getscansummary()
    mstool.close()
    scans = np.sort(np.array([int(i) for i in scan_summary.keys()]))
    target_scans, cal_scans, f_scans, g_scans, p_scans = get_cal_target_scans(msname)
    selected_field = []
    valid_scans = []
    if field != "":
        field = field.split(",")
        msmd = msmetadata()
        msmd.open(msname)
        for f in field:
            with suppress_casa_output():
                try:
                    field_id = msmd.fieldsforname(f)[0]
                except Exception as e:
                    field_id = int(f)
            selected_field.append(field_id)
        msmd.close()
        msmd.done()
        del msmd
    for scan in scans:
        scan_field = scan_summary[str(scan)]["0"]["FieldId"]
        if len(selected_field) == 0 or scan_field in selected_field:
            duration = (
                scan_summary[str(scan)]["0"]["EndTime"]
                - scan_summary[str(scan)]["0"]["BeginTime"]
            ) * 86400.0
            duration = round(duration / 60.0, 1)
            if duration >= min_scan_time:
                valid_scans.append(scan)
    return valid_scans


def get_target_fields(msname):
    """
    Get target fields

    Parameters
    ----------
    msname : str
        Name of the measurement set

    Returns
    -------
    list
        Target field names
    dict
        Target field scans
    """
    fluxcal_field, fluxcal_scans = get_fluxcals(msname)
    phasecal_field, phasecal_scans, phasecal_fluxs = get_phasecals(msname)
    calibrator_field = fluxcal_field + phasecal_field
    msmd = msmetadata()
    msmd.open(msname)
    field_names = msmd.fieldnames()
    field_names = np.unique(field_names)
    target_fields = []
    target_scans = {}
    for f in field_names:
        if f not in calibrator_field:
            target_fields.append(f)
    for field in target_fields:
        scans = msmd.scansforfield(field).tolist()
        target_scans[field] = scans
    msmd.close()
    msmd.done()
    del msmd
    return target_fields, target_scans


def get_caltable_fields(caltable):
    """
    Get caltable field names

    Parameters
    ----------
    caltable : str
        Caltable name

    Returns
    -------
    list
        Field names
    """
    tb = table()
    tb.open(caltable + "/FIELD")
    field_names = tb.getcol("NAME")
    field_ids = tb.getcol("SOURCE_ID")
    tb.close()
    tb.open(caltable)
    fields = np.unique(tb.getcol("FIELD_ID"))
    tb.close()
    field_name_list = []
    for f in fields:
        pos = np.where(field_ids == f)[0][0]
        field_name_list.append(str(field_names[pos]))
    return field_name_list


def get_cal_target_scans(msname):
    """
    Get calibrator and target scans

    Parameters
    ----------
    msname : str
        Name of the measurement set

    Returns
    -------
    list
        Target scan numbers
    list
        Calibrator scan numbers
    list
        Fluxcal scans
    list
        Phasecal scans
    list
        Polcal scans
    """
    f_scans = []
    p_scans = []
    g_scans = []
    fluxcal_fields, fluxcal_scans = get_fluxcals(msname)
    phasecal_fields, phasecal_scans, phasecal_flux_list = get_phasecals(msname)
    polcal_fields, polcal_scans = get_polcals(msname)
    for fluxcal_scan in fluxcal_scans.values():
        for s in fluxcal_scan:
            f_scans.append(s)
    for polcal_scan in polcal_scans.values():
        for s in polcal_scan:
            p_scans.append(s)
    for phasecal_scan in phasecal_scans.values():
        for s in phasecal_scan:
            g_scans.append(s)
    cal_scans = f_scans + p_scans + g_scans
    msmd = msmetadata()
    msmd.open(msname)
    all_scans = msmd.scannumbers()
    msmd.close()
    msmd.done()
    target_scans = []
    for scan in all_scans:
        if scan not in cal_scans:
            target_scans.append(scan)
    return target_scans, cal_scans, f_scans, g_scans, p_scans


def get_band_name(msname):
    """
    Get band name

    Parameters
    ----------
    msname : str
        Name of the ms

    Returns
    -------
    str
        Band name ('U','L','S')
    """
    msmd = msmetadata()
    msmd.open(msname)
    meanfreq = msmd.meanfreq(0) / 10**6
    msmd.close()
    msmd.done()
    if meanfreq >= 544 and meanfreq <= 1088:
        return "U"
    elif meanfreq >= 856 and meanfreq <= 1712:
        return "L"
    else:
        return "S"


def get_bad_chans(msname):
    """
    Get bad channels to flag

    Parameters
    ----------
    msname : str
        Name of the ms

    Returns
    -------
    str
        SPW string of bad channels
    """
    msmd = msmetadata()
    msmd.open(msname)
    chanfreqs = msmd.chanfreqs(0) / 10**6
    msmd.close()
    msmd.done()
    bandname = get_band_name(msname)
    if bandname == "U":
        bad_freqs = [
            (-1, 580),
            (925, 960),
            (1010, -1),
        ]
    elif bandname == "L":
        bad_freqs = [
            (-1, 879),
            (925, 960),
            (1166, 1186),
            (1217, 1237),
            (1242, 1249),
            (1375, 1387),
            (1526, 1626),
            (1681, -1),
        ]
    else:
        print("MeerKAT data is not in UHF or L-band.")
        bad_freqs = []
    if min(chanfreqs) <= bad_freqs[0][1] and max(chanfreqs) >= bad_freqs[-1][0]:
        spw = "0:"
        count = 0
        for freq_range in bad_freqs:
            start_freq = freq_range[0]
            end_freq = freq_range[1]
            if start_freq == -1:
                start_chan = 0
            else:
                start_chan = np.argmin(np.abs(start_freq - chanfreqs))
            if count > 0 and start_chan <= end_chan:
                break
            if end_freq == -1:
                end_chan = len(chanfreqs) - 1
            else:
                end_chan = np.argmin(np.abs(end_freq - chanfreqs))
            if end_chan > start_chan:
                spw += str(start_chan) + "~" + str(end_chan) + ";"
            else:
                spw += str(start_chan) + ";"
            count += 1
        spw = spw[:-1]
    else:
        spw = ""
    return spw


def get_good_chans(msname):
    """
    Get good channel range to perform gaincal

    Parameters
    ----------
    msname : str
        Name of the ms

    Returns
    -------
    str
        SPW string
    """
    msmd = msmetadata()
    msmd.open(msname)
    chanfreqs = msmd.chanfreqs(0) / 10**6
    meanfreq = msmd.meanfreq(0) / 10**6
    msmd.close()
    msmd.done()
    bandname = get_band_name(msname)
    if bandname == "U":
        good_freqs = [(580, 620)]  # For UHF band
    elif bandname == "L":
        good_freqs = [(890, 920)]  # For L band
    else:
        good_freqs = []  # For S band
    if min(chanfreqs) <= good_freqs[0][1] and max(chanfreqs) >= good_freqs[-1][0]:
        spw = "0:"
        for freq_range in good_freqs:
            start_freq = freq_range[0]
            end_freq = freq_range[1]
            start_chan = np.argmin(np.abs(start_freq - chanfreqs))
            end_chan = np.argmin(np.abs(end_freq - chanfreqs))
            spw += str(start_chan) + "~" + str(end_chan) + ";"
        spw = spw[:-1]
    else:
        spw = f"0:0~{len(chanfreqs)-1}"
    return spw


###########################
# Spliting noise diode
###########################


def split_noise_diode_scans(
    msname="",
    noise_on_ms="",
    noise_off_ms="",
    field="",
    scan="",
    datacolumn="data",
    n_threads=-1,
    dry_run=True,
):
    """
    Split noise diode on and off timestamps into two seperate measurement sets

    Parameters
    ----------
    msname : str
        Measurement set
    noise_on_ms : str, optional
        Noise diode on ms
    noise_off_ms : str, optional
        Noise diode off ms
    field : str, optional
        Field name or id
    scan : str, optional
        Scan number
    datacolumn : str, optional
        Data column to split

    Returns
    -------
    tuple
        splited ms names
    """
    limit_threads(n_threads=n_threads)
    from casatasks import split

    if dry_run:
        process = psutil.Process(os.getpid())
        mem = round(process.memory_info().rss / 1024**3, 2)  # in GB
        return mem
    msname = msname.rstrip("/")
    mspath = os.path.dirname(os.path.abspath(msname))
    os.chdir(mspath)
    print(f"Spliting ms: {msname} into noise diode on and off measurement sets.")
    if noise_on_ms == "":
        noise_on_ms = msname.split(".ms")[0] + "_noise_on.ms"
    if noise_off_ms == "":
        noise_off_ms = msname.split(".ms")[0] + "_noise_off.ms"
    if os.path.exists(noise_on_ms):
        os.system("rm -rf " + noise_on_ms)
    if os.path.exists(noise_on_ms + ".flagversions"):
        os.system("rm -rf " + noise_on_ms + ".flagversions")
    if os.path.exists(noise_off_ms):
        os.system("rm -rf " + noise_off_ms)
    if os.path.exists(noise_off_ms + ".flagversions"):
        os.system("rm -rf " + noise_off_ms + ".flagversions")
    tb = table()
    tb.open(msname)
    times = tb.getcol("TIME")
    tb.close()
    unique_times = np.unique(times)
    even_times = unique_times[::2]  # Even-indexed timestamps
    odd_times = unique_times[1::2]  # Odd-indexed timestamps
    even_timerange = ",".join(
        [mjdsec_to_timestamp(t, str_format=1) for t in even_times]
    )
    odd_timerange = ",".join([mjdsec_to_timestamp(t, str_format=1) for t in odd_times])
    even_ms = msname.split(".ms")[0] + "_even.ms"
    odd_ms = msname.split(".ms")[0] + "_odd.ms"
    split(
        vis=msname,
        outputvis=even_ms,
        timerange=even_timerange,
        field=field,
        scan=scan,
        datacolumn=datacolumn,
    )
    split(
        vis=msname,
        outputvis=odd_ms,
        timerange=odd_timerange,
        field=field,
        scan=scan,
        datacolumn=datacolumn,
    )
    mstool = casamstool()
    mstool.open(even_ms)
    mstool.select({"antenna1": 1, "antenna2": 1})
    even_data = np.nanmean(np.abs(mstool.getdata("DATA")["data"]))
    mstool.close()
    mstool.open(odd_ms)
    mstool.select({"antenna1": 1, "antenna2": 1})
    odd_data = np.nanmean(np.abs(mstool.getdata("DATA")["data"]))
    mstool.close()
    if even_data > odd_data:
        os.system("mv " + even_ms + " " + noise_on_ms)
        os.system("mv " + odd_ms + " " + noise_off_ms)
    else:
        os.system("mv " + odd_ms + " " + noise_on_ms)
        os.system("mv " + even_ms + " " + noise_off_ms)
    return noise_on_ms, noise_off_ms


def determine_noise_diode_cal_scan(msname, scan):
    """
    Determine whether a calibrator scan is a noise-diode cal scan or not

    Parameters
    ----------
    msname : str
        Name of the measurement set
    scan : int
        Scan number

    Returns
    -------
    bool
        Whether it is noise-diode cal scan or not
    """

    def is_noisescan(msname, chan, scan):
        mstool = casamstool()
        mstool.open(msname)
        mstool.select({"antenna1": 1, "antenna2": 1, "scan_number": scan})
        mstool.selectchannel(nchan=1, width=1, start=chan)
        data = mstool.getdata("DATA", ifraxis=True)["data"][:, 0, 0, :]
        mstool.close()
        xx = np.abs(data[0, ...])
        yy = np.abs(data[-1, ...])
        even_xx = xx[1::2]
        odd_xx = xx[::2]
        minlen = min(len(even_xx), len(odd_xx))
        d_xx = even_xx[:minlen] - odd_xx[:minlen]
        even_yy = yy[1::2]
        odd_yy = yy[::2]
        d_yy = even_yy[:minlen] - odd_yy[:minlen]
        mean_d_xx = np.abs(np.nanmedian(d_xx))
        mean_d_yy = np.abs(np.nanmedian(d_yy))
        if mean_d_xx > 10 and mean_d_yy > 10:
            return True
        else:
            return False

    print(f"Check noise-diode cal for scan : {scan}")
    good_spw = get_good_chans(msname)
    chan = int(good_spw.split(";")[0].split(":")[-1].split("~")[0])
    return is_noisescan(msname, chan, scan)



import argparse
import sys
import os
from casatasks import casalog
from casatasks import listobs
from meersolar.utils.resource_utils import drop_cache

try:
    casalogfile = casalog.logfile()
    os.system("rm -rf " + casalogfile)
except BaseException:
    pass


def show_listobs(msname):
    if not os.path.exists(msname):
        raise FileNotFoundError(f"Measurement Set not found: {msname}")
    listfile = msname.split(".ms")[0] + ".listobs"
    os.system(f"rm -rf {listfile}")
    listobs(vis=msname, listfile=listfile, verbose=True)
    with open(listfile, "r") as f:
        lines = f.readlines()
    filtered_lines = []
    for line in lines:
        if "Sources:" in line:
            break
        filtered_lines.append(line)
    os.system(f"rm -rf {listfile}")
    print("".join(filtered_lines))
    drop_cache(msname)


def cli():
    parser = argparse.ArgumentParser(description="Run listobs and show from saved file")
    parser.add_argument("msname", type=str, help="Path to the measurement set")
    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)
    args = parser.parse_args()
    show_listobs(args.msname)


if __name__ == "__main__":
    cli()



import astropy.units as u
import logging
import dask
import numpy as np
import argparse
import traceback
import warnings
import copy
import time
import sys
import os
from astropy.io import fits
from astropy.time import Time
from astropy.coordinates import EarthLocation, SkyCoord, AltAz
from astropy.wcs import FITSFixedWarning
from casatasks import casalog
from numpy.linalg import inv
from astropy.wcs import WCS
from scipy.interpolate import RectBivariateSpline
from joblib import Parallel, delayed as joblid_delayed
from meersolar.utils import (
    get_datadir,
    get_cachedir,
    SmartDefaultsHelpFormatter,
    save_pid,
)

logging.getLogger("distributed").setLevel(logging.WARNING)

try:
    casalogfile = casalog.logfile()
    os.system("rm -rf " + casalogfile)
except BaseException:
    pass

datadir = get_datadir()

warnings.simplefilter("ignore", category=FITSFixedWarning)

# Define MeerKAT location
MEERLAT = -30.7133
MEERLON = 21.4429
MEERALT = 1086.6
datadir = get_datadir()


def get_IQUV(filename):
    """
    Get IQUV from a fits

    Parameters
    ----------
    filename : str
        Fits image name

    Returns
    -------
    dict
        Stokes
    """
    data = fits.getdata(filename).astype("float32")
    header = fits.getheader(filename)
    keys = header.keys()
    if "CTYPE3" in keys and header["CTYPE3"] == "STOKES":
        stokesaxis = 3
    elif "CTYPE4" in keys and header["CTYPE4"] == "STOKES":
        stokesaxis = 4
    else:
        stokesaxis = 1
    shape = data.shape
    stokes = {}
    if shape[0] > 1 or shape[1] > 1 and (stokesaxis == 3 or stokesaxis == 4):
        if stokesaxis == 3:
            stokes["I"] = data[0, 0, :, :]
            stokes["Q"] = data[0, 1, :, :]
            stokes["U"] = data[0, 2, :, :]
            stokes["V"] = data[0, 3, :, :]
        elif stokesaxis == 4:
            stokes["I"] = data[0, 0, :, :]
            stokes["Q"] = data[1, 0, :, :]
            stokes["U"] = data[2, 0, :, :]
            stokes["V"] = data[3, 0, :, :]
    else:
        stokes["I"] = data[0, 0, :, :]
        stokes["Q"] = copy.deepcopy(stokes["I"]) * 0
        stokes["U"] = copy.deepcopy(stokes["I"]) * 0
        stokes["V"] = copy.deepcopy(stokes["I"]) * 0
    return stokes


def put_IQUV(filename, stokes, header):
    """
    Put IQUV into a fits

    Parameters
    ----------
    filename : str
        Fits image name
    stokes : dict
        Stokes
    header : dict
        Image header

    Returns
    -------
    dict
        Stokes
    """
    keys = header.keys()
    if "CTYPE3" in keys and header["CTYPE3"] == "STOKES":
        stokesaxis = 3
    elif "CTYPE4" in keys and header["CTYPE4"] == "STOKES":
        stokesaxis = 4
    else:
        stokesaxis = 1
    naxis = header["NAXIS"]
    shape = tuple(header[f"NAXIS{axis}"] for axis in range(naxis, 0, -1))
    data = np.empty(shape, dtype=np.float32)
    if shape[0] > 1 or shape[1] > 1 and (stokesaxis == 3 or stokesaxis == 4):
        if stokesaxis == 3:
            data[0, 0, :, :] = stokes["I"]
            data[0, 1, :, :] = stokes["Q"]
            data[0, 2, :, :] = stokes["U"]
            data[0, 3, :, :] = stokes["V"]
        elif stokesaxis == 4:
            data[0, 0, :, :] = stokes["I"]
            data[1, 0, :, :] = stokes["Q"]
            data[2, 0, :, :] = stokes["U"]
            data[3, 0, :, :] = stokes["V"]
    else:
        data[0, 0, :, :] = stokes["I"]
    fits.writeto(filename, data=data, header=header, overwrite=True)
    return filename


def get_brightness(stokes):
    """

    Returns brightness matrix from stokes dictionary (X and Y are in opposite convention of IAU in MeerKAT)
    """
    I = stokes["I"].astype("float32")
    Q = stokes["Q"].astype("float32")
    U = stokes["U"].astype("float32")
    V = stokes["V"].astype("float32")
    XX = I - Q
    XY = U - 1j * V
    YX = U + 1j * V
    YY = I + Q
    B = np.array([XX, XY, YX, YY]).astype("complex64")
    B = B.T
    B = B.reshape(B.shape[0], B.shape[1], 2, 2)
    return B


def make_stokes(b):
    """
    Makes stokes images from brightness matrix
    """
    XX = b[0, 0, ...].astype("complex64")
    XY = b[0, 1, ...].astype("complex64")
    YX = b[1, 0, ...].astype("complex64")
    YY = b[1, 1, ...].astype("complex64")
    stokes = {}
    stokes["I"] = np.real(XX + YY) / 2.0
    stokes["Q"] = np.real(YY - XX) / 2.0
    stokes["U"] = np.real(XY + YX) / 2.0
    stokes["V"] = np.real(1j * (XY - YX)) / 2.0
    return stokes


def load_beam(image_file, band=""):
    """
    Load MeerKAT beam

    Parameters
    ----------
    image_file : str
        Image name (Assuming single spectral image)
    band : str, optional
        Band name (If not provided, check from header or frequency)

    Returns
    -------
    numpy.array
        l,m coordinates
    numpy.array
        Full Jones complex beam
    """
    hdr = fits.getheader(image_file)
    keys = hdr.keys()
    if "CTYPE3" in keys and hdr["CTYPE3"] == "FREQ":
        freq = hdr["CRVAL3"]
        delfreq = hdr["CDELT3"]
    elif "CTYPE4" in keys and hdr["CTYPE4"] == "FREQ":
        freq = hdr["CRVAL4"]
        delfreq = hdr["CDELT4"]
    else:
        print(f"No frequency axis in image {image_file}.")
        return
    freq1 = (freq - (delfreq / 2)) / 10**6  # In MHz
    freq2 = (freq + (delfreq / 2)) / 10**6  # In MHz
    if band == "":
        try:
            band = hdr["BAND"]
        except BaseException:
            if freq1 >= 544 and freq2 <= 1088:  # UHF band
                band = "U"
            elif freq1 >= 856 and freq2 <= 1712:  # L band
                band = "L"
            else:
                print(f"Image: {image_file} is not in UHF or L-band.")
                return
    if band == "U":
        beam_data = np.load(datadir + "/MeerKAT_antavg_Uband.npz", mmap_mode="r")
    elif band == "L":
        beam_data = np.load(datadir + "/MeerKAT_antavg_Lband.npz", mmap_mode="r")
    else:
        print(f"Image: {image_file} is not in UHF or L-band.")
        return
    freqs = beam_data["freqs"]
    coords = np.deg2rad(
        beam_data["coords"]
    )  # It is done as l,m values were converted into degree
    pos = np.where((freq >= freq1) & (freqs <= freq1))[0]
    beam = beam_data["beams"][:, pos, ...].mean(1)
    beam = beam.astype("complex64")
    del beam_data, freqs
    return coords, beam


def get_radec_grid(image_file):
    """
    Get RA and Dec arrays for all pixels in an image.

    Parameters
    ----------
    image_file : str
        FITS image file name

    Returns
    -------
    ra : 2D numpy.ndarray
        RA values in degrees for each pixel
    dec : 2D numpy.ndarray
        Dec values in degrees for each pixel
    """
    hdr = fits.getheader(image_file)
    wcs = WCS(hdr).celestial
    ny, nx = hdr["NAXIS2"], hdr["NAXIS1"]
    y, x = np.mgrid[0:ny, 0:nx]  # pixel coordinates
    world = wcs.pixel_to_world(x, y)
    ra = world.ra.deg
    dec = world.dec.deg
    return ra, dec


def get_pointingcenter_radec(image_file):
    """
    Get image pointing center RA DEC

    Parameters
    ----------
    image_file : str
        Image file name

    Returns
    -------
    float
        RA in degree
    float
        DEC in degree
    """
    hdr = fits.getheader(image_file)
    image_wcs = WCS(hdr)
    image_shape = (hdr["NAXIS2"], hdr["NAXIS1"])
    ra0 = float(hdr["CRVAL1"])
    dec0 = float(hdr["CRVAL2"])
    return ra0, dec0


def radec_to_lm(ra_deg, dec_deg, ra0_deg, dec0_deg):
    """
    Convert RA/Dec to l,m direction cosines relative to a phase center.

    Parameters
    ----------
    ra_deg, dec_deg : 2D arrays
        RA and Dec in degrees
    ra0_deg, dec0_deg : float
        Phase center RA and Dec in degrees

    Returns
    -------
    l, m : 2D arrays
        Direction cosines (dimensionless)
    """
    ra = np.radians(ra_deg)
    dec = np.radians(dec_deg)
    ra0 = np.radians(ra0_deg)
    dec0 = np.radians(dec0_deg)
    delta_ra = ra - ra0
    l = np.cos(dec) * np.sin(delta_ra)
    m = np.sin(dec) * np.cos(dec0) - np.cos(dec) * np.sin(dec0) * np.cos(delta_ra)
    return l, m


def get_parallactic_angle(
    obs_time, ra_deg, dec_deg, LAT=MEERLAT, LON=MEERLON, ALT=MEERALT
):
    """
    Get parallactic angle

    Parameters
    ----------
    obs_time : str
        Observation time in YYY-MM-DDThh:mm:ss format
    ra : float
        RA in degree
    dec : float
        DEC in degree

    Returns
    -------
    float
        Parallactic angle in degree
    """
    sky = SkyCoord(ra=ra_deg * u.deg, dec=dec_deg * u.deg, frame="icrs")
    obstime = Time(obs_time)
    meerpos = EarthLocation(lat=LAT * u.deg, lon=LON * u.deg, height=ALT * u.m)
    altaz = sky.transform_to(AltAz(obstime=obstime, location=meerpos))
    az = altaz.az.rad
    alt = altaz.alt.rad
    lat = np.deg2rad(LAT)
    p = np.arctan2(
        np.sin(az) * np.cos(lat),
        np.cos(alt) * np.sin(lat) - np.sin(alt) * np.cos(lat) * np.cos(az),
    )
    return round(np.rad2deg(p), 2)


def get_beam_interpolator(jones, coords):
    """
    Get beam interpolator

    Parameters
    ----------
    jones : numpy.array
        Jones array (shape, npol, l_npix, m_npix)
    coords : numpy.array
        l,m coordinates

    Returns
    -------
    interpolator
        Interpolation functions
    """
    j00_r = RectBivariateSpline(
        x=coords, y=coords, z=np.nan_to_num(np.real(jones[0, ...]))
    )
    j00_i = RectBivariateSpline(
        x=coords, y=coords, z=np.nan_to_num(np.imag(jones[0, ...]))
    )
    j01_r = RectBivariateSpline(
        x=coords, y=coords, z=np.nan_to_num(np.real(jones[1, ...]))
    )
    j01_i = RectBivariateSpline(
        x=coords, y=coords, z=np.nan_to_num(np.imag(jones[1, ...]))
    )
    j10_r = RectBivariateSpline(
        x=coords, y=coords, z=np.nan_to_num(np.real(jones[2, ...]))
    )
    j10_i = RectBivariateSpline(
        x=coords, y=coords, z=np.nan_to_num(np.imag(jones[2, ...]))
    )
    j11_r = RectBivariateSpline(
        x=coords, y=coords, z=np.nan_to_num(np.real(jones[3, ...]))
    )
    j11_i = RectBivariateSpline(
        x=coords, y=coords, z=np.nan_to_num(np.imag(jones[3, ...]))
    )
    return j00_r, j00_i, j01_r, j01_i, j10_r, j10_i, j11_r, j11_i


def apply_parallactic_rotation(jones, p_angle):
    """
    Apply left-side parallactic rotation: J' = J.R(p_angle)
    as needed in the RIME context (sky-frame transformation).

    Parameters
    ----------
    jones : ndarray
        Jones matrix, shape (4, H, W), with components:
        [0] = J_00, [1] = J_01, [2] = J_10, [3] = J_11
    chi : float
        Parallactic angle in degree

    Returns
    -------
    jones_rot : ndarray
        Rotated Jones matrix, shape (4, H, W)
    """
    p_angle = np.deg2rad(p_angle)
    c = np.cos(p_angle)
    s = -np.sin(p_angle)
    j00, j01, j10, j11 = jones
    jj00 = j00 * c - j01 * s
    jj01 = j00 * s + j01 * c
    jj10 = j10 * c - j11 * s
    jj11 = j10 * s + j11 * c
    return np.stack([jj00, jj01, jj10, jj11], axis=0).astype("complex64")


def get_image_beam(
    image_file,
    pbdir,
    save_beam=True,
    band="",
    apply_parang=True,
    n_cpu=8,
    verbose=False,
):
    """
    Get image beam

    Parameters
    ----------
    image_file : str
        Image file name
    pbdir : str
        Primary beam directory
    save_beam : bool, optional
        Save beam of the image
    band : str, optional
        Band name
    apply_parang : bool, optional
        Apply parallactic angle correction
    n_cpu : int, optinal
        Number of CPU threads to use
    verbose : bool, optional
        Verbose output

    Returns
    -------
    numpy.array
        Jones array
    """
    if n_cpu > 8:
        n_cpu = 8
    start_time = time.time()
    ##################################
    header = fits.getheader(image_file)
    if header["CTYPE3"] == "FREQ":
        freq = header["CRVAL3"]
    elif header["CTYPE4"] == "FREQ":
        freq = header["CRVAL4"]
    else:
        print(f"No frequency axis in image: {image_file}.")
        return
    freq = round(freq / 10**6, 1)  # In MHz
    pbfile = f"{pbdir}/freq_{freq}_pb.npy"
    obs_time = header["DATE-OBS"]
    ra0, dec0 = get_pointingcenter_radec(image_file)  # Phase center
    p_angle = get_parallactic_angle(
        obs_time, ra0, dec0
    )  # Parallactic angle of the center
    #######################################
    # If beam file exists
    #######################################
    fresh_run = True
    if os.path.exists(pbfile):
        if verbose:
            print(f"Loading beam from: {pbfile}")
        try:
            jones_array = np.load(pbfile, allow_pickle=True)
            fresh_run = False
        except BaseException:
            fresh_run = True
            os.system(f"rm -rf {pbfile}")
    #################################
    # Fresh run
    #################################
    if fresh_run:
        ra_grid, dec_grid = get_radec_grid(image_file)  # RA DEC grid
        l_grid, m_grid = radec_to_lm(ra_grid, dec_grid, ra0, dec0)
        ############################
        # Load beam
        ############################
        beam_results = load_beam(image_file, band=band)
        if beam_results is None:
            return
        lm_coords, beam = beam_results
        j00_r, j00_i, j01_r, j01_i, j10_r, j10_i, j11_r, j11_i = get_beam_interpolator(
            beam, lm_coords
        )
        l_grid_flat = l_grid.flatten()
        m_grid_flat = m_grid.flatten()
        grid_shape = l_grid.shape
        del l_grid, m_grid
        with Parallel(n_jobs=n_cpu, backend="threading") as parallel:
            results = parallel(
                [
                    joblid_delayed(j00_r)(l_grid_flat, m_grid_flat, grid=False),
                    joblid_delayed(j00_i)(l_grid_flat, m_grid_flat, grid=False),
                    joblid_delayed(j01_r)(l_grid_flat, m_grid_flat, grid=False),
                    joblid_delayed(j01_i)(l_grid_flat, m_grid_flat, grid=False),
                    joblid_delayed(j10_r)(l_grid_flat, m_grid_flat, grid=False),
                    joblid_delayed(j10_i)(l_grid_flat, m_grid_flat, grid=False),
                    joblid_delayed(j11_r)(l_grid_flat, m_grid_flat, grid=False),
                    joblid_delayed(j11_i)(l_grid_flat, m_grid_flat, grid=False),
                ]
            )
        del parallel
        (
            j00_r_arr,
            j00_i_arr,
            j01_r_arr,
            j01_i_arr,
            j10_r_arr,
            j10_i_arr,
            j11_r_arr,
            j11_i_arr,
        ) = results
        j00_r_arr = j00_r_arr.reshape(grid_shape)
        j00_i_arr = j00_i_arr.reshape(grid_shape)
        j01_r_arr = j01_r_arr.reshape(grid_shape)
        j01_i_arr = j01_i_arr.reshape(grid_shape)
        j10_r_arr = j10_r_arr.reshape(grid_shape)
        j10_i_arr = j10_i_arr.reshape(grid_shape)
        j11_r_arr = j11_r_arr.reshape(grid_shape)
        j11_i_arr = j11_i_arr.reshape(grid_shape)
        jones_array = np.array(
            [
                j00_r_arr + 1j * j00_i_arr,
                j01_r_arr + 1j * j01_i_arr,
                j10_r_arr + 1j * j10_i_arr,
                j11_r_arr + 1j * j11_i_arr,
            ]
        ).astype("complex64")
        del (
            j00_r_arr,
            j00_i_arr,
            j01_r_arr,
            j01_i_arr,
            j10_r_arr,
            j10_i_arr,
            j11_r_arr,
            j11_i_arr,
        )
        if save_beam and os.path.exists(pbfile) == False:
            np.save(pbfile, np.array(jones_array, dtype="object"))
            if verbose:
                print(f"Beam saved in: {pbfile}")
    if apply_parang:
        # This is to account B'=P(X)BP(-X) parallactic trasnform on brightness
        # matrix
        jones_array = apply_parallactic_rotation(jones_array, p_angle).T
    jones_array = jones_array.reshape(jones_array.shape[0], jones_array.shape[1], 2, 2)
    if verbose:
        print(f"Beam calculated in : {round(time.time()-start_time,1)}s")
    return jones_array


def get_pbcor_image(
    image_file,
    pbdir,
    pbcor_dir,
    save_beam=True,
    band="",
    apply_parang=True,
    n_cpu=8,
    verbose=False,
):
    """
    Get primary beam corrected image

    Parameters
    ----------
    image_file : str
        Image file name
    pbdir : str
        Primary beam directory
    pbcor_dir : str
        Primary beam corrected image directory
    save_beam : bool, optional
        Save the beam for the image
    band : str, optional
        Band name
    apply_parang : bool, optional
        Apply parallactic correction
    n_cpu : int, optional
        Number of CPU threads to use
    verbose : bool, optional
        Verbose output

    Returns
    -------
    str
        Primary beam corrected image
    """
    try:
        image_file = image_file.rstrip("/")
        print(f"Correcting beam for image: {os.path.basename(image_file)}...")
        beam = get_image_beam(
            image_file,
            pbdir,
            save_beam=save_beam,
            band=band,
            apply_parang=apply_parang,
            n_cpu=int(n_cpu),
            verbose=verbose,
        )
        if not isinstance(beam, np.ndarray):
            print(f"Error in correct beam for image: {os.path.basename(image_file)}")
            return
        det = beam[..., 0, 0] * beam[..., 1, 1] - beam[..., 0, 1] * beam[..., 1, 0]
        inv_beam = np.empty_like(beam, dtype=np.complex64)
        inv_beam[..., 0, 0] = beam[..., 1, 1] / det
        inv_beam[..., 0, 1] = -beam[..., 0, 1] / det
        inv_beam[..., 1, 0] = -beam[..., 1, 0] / det
        inv_beam[..., 1, 1] = beam[..., 0, 0] / det
        beam_H = np.conj(np.swapaxes(beam, -1, -2))
        del beam
        det = (
            beam_H[..., 0, 0] * beam_H[..., 1, 1]
            - beam_H[..., 0, 1] * beam_H[..., 1, 0]
        )
        inv_beam_H = np.empty_like(beam_H, dtype=np.complex64)
        inv_beam_H[..., 0, 0] = beam_H[..., 1, 1] / det
        inv_beam_H[..., 0, 1] = -beam_H[..., 0, 1] / det
        inv_beam_H[..., 1, 0] = -beam_H[..., 1, 0] / det
        inv_beam_H[..., 1, 1] = beam_H[..., 0, 0] / det
        del beam_H
        image_stokes = get_IQUV(image_file)
        B_matrix = get_brightness(image_stokes)
        del image_stokes
        B_tmp = np.matmul(B_matrix, inv_beam_H)
        del inv_beam_H
        B_cor = np.matmul(inv_beam, B_tmp)
        del B_tmp, inv_beam
        B_cor = np.transpose(B_cor, (2, 3, 1, 0))
        pbcor_stokes = make_stokes(B_cor)
        del B_cor
        #################################
        pbcor_file = (
            pbcor_dir
            + "/"
            + os.path.basename(image_file).split(".fits")[0]
            + "_pbcor.fits"
        )
        header = fits.getheader(image_file)
        pbcor_file = put_IQUV(pbcor_file, pbcor_stokes, header)
        return pbcor_file
    except Exception as e:
        traceback.print_exc()
        return


def main(
    imagename,
    pbdir="",
    pbcor_dir="",
    save_beam=True,
    band="",
    apply_parang=True,
    verbose=False,
    ncpu=1,
    jobid=0,
):
    """
    Single image primary beam correction for MeerKAT

    Parameters
    ----------
    imagename : str
        Image name
    pbdir : str, optional
        Primary beam directory
    pbcor_dir : str, optional
        Primary beam corrected image directory
    save_beam : bool, optional
        Save primary beams for later use
    band : str, optional
        MeerKAT band name
    apply_parang : bool, optional
        Applt parallactic correction
    verbose : bool, optional
        Verbose output
    ncpu : int, optional
        Number of CPUs to use
    jobid : str, optional
        Job ID

    Returns
    -------
    int
        Success message
    """
    pid = os.getpid()
    cachedir = get_cachedir()
    save_pid(pid, f"{cachedir}/pids/pids_{jobid}.txt")

    try:
        if imagename and os.path.exists(imagename):
            if pbdir == "":
                print("Provide an existing directory name in pbdir.")
                msg = 1
            else:
                os.makedirs(pbdir, exist_ok=True)
                if pbcor_dir == "":
                    pbcor_dir = pbdir
                os.makedirs(pbcor_dir, exist_ok=True)
                pbcor_image = get_pbcor_image(
                    imagename,
                    pbdir,
                    pbcor_dir,
                    band=band,
                    apply_parang=apply_parang,
                    save_beam=save_beam,
                    n_cpu=int(ncpu),
                    verbose=verbose,
                )
                if pbcor_image is None or not os.path.exists(pbcor_image):
                    msg = 1
                    print(f"Primary beam correction is not successful")
                else:
                    msg = 0
                    print(f"Primary beam corrected image: {pbcor_image}")
        else:
            print("Please provide correct image name.")
            msg = 1
    except Exception as e:
        traceback.print_exc()
        msg = 1
    return msg


def cli():
    parser = argparse.ArgumentParser(
        description="Correct image for full-polar antenna averaged MeerKAT primary beam",
        formatter_class=SmartDefaultsHelpFormatter,
    )

    # Essential parameters
    basic_args = parser.add_argument_group(
        "###################\nEssential parameters\n###################"
    )
    basic_args.add_argument(
        "imagename", type=str, help="Name of image (required positional argument)"
    )
    basic_args.add_argument(
        "--pbdir",
        type=str,
        default="",
        help="Name of primary beam directory",
    )
    basic_args.add_argument(
        "--pbcor_dir",
        type=str,
        default="",
        help="Name of primary beam corrected image directory",
    )

    # Advanced parameters
    adv_args = parser.add_argument_group(
        "###################\nAdvanced parameters\n###################"
    )
    adv_args.add_argument(
        "--no_save_beam",
        action="store_false",
        dest="save_beam",
        help="Do not save beam to disk",
    )
    adv_args.add_argument("--band", type=str, default="", help="Band name")
    adv_args.add_argument(
        "--no_apply_parang",
        action="store_false",
        dest="apply_parang",
        help="Do not apply parallactic angle correction",
    )
    adv_args.add_argument("--verbose", action="store_true", help="Verbose output")

    # Resource management parameters
    hard_args = parser.add_argument_group(
        "###################\nHardware resource management parameters\n###################"
    )
    hard_args.add_argument(
        "--ncpu",
        type=int,
        default=1,
        help="Number of CPU threads to use",
        metavar="Integer",
    )
    hard_args.add_argument("--jobid", type=int, default=0, help="Job ID")

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    args = parser.parse_args()

    msg = main(
        imagename=args.imagename,
        pbdir=args.pbdir,
        pbcor_dir=args.pbcor_dir,
        save_beam=args.save_beam,
        band=args.band,
        apply_parang=args.apply_parang,
        verbose=args.verbose,
        ncpu=args.ncpu,
        jobid=args.jobid,
    )
    return msg


if __name__ == "__main__":
    result = cli()
    os._exit(result)



import types
import astropy.units as u
import logging
import psutil
import numpy as np
import warnings
import glob
import dask
import os
import traceback
import matplotlib
import matplotlib.pyplot as plt
from dask import delayed, compute
from multiprocessing.pool import ThreadPool
from sunpy.net import Fido, attrs as a
from sunpy.map import Map
from astropy.visualization import ImageNormalize, PowerStretch, LogStretch
from astropy.io import fits
from astropy.time import Time
from astropy.coordinates import EarthLocation, SkyCoord
from astropy.wcs import FITSFixedWarning
from casatools import msmetadata, ms as casamstool
from datetime import datetime as dt, timedelta
from .basic_utils import *
from .proc_manage_utils import *
from .ms_metadata import *
from .meer_utils import *

warnings.simplefilter("ignore", category=FITSFixedWarning)


#################################
# Plotting related functions
#################################


def get_meermap(fits_image, band="", do_sharpen=False):
    """
    Make MeerKAT sunpy map

    Parameters
    ----------
    fits_image : str
        MeerKAT fits image
    band : str, optional
        Band name
    do_sharpen : bool, optional
        Sharpen the image

    Returns
    -------
    sunpy.map
        Sunpy map
    """
    from scipy.ndimage import gaussian_filter
    from sunpy.map import make_fitswcs_header
    from sunpy.coordinates import frames, sun

    logging.getLogger("sunpy").setLevel(logging.ERROR)

    MEERLAT = -30.7133
    MEERLON = 21.4429
    MEERALT = 1086.6
    meer_hdu = fits.open(fits_image)  # Opening MeerKAT fits file
    meer_header = meer_hdu[0].header  # meer header
    meer_data = meer_hdu[0].data
    if len(meer_data.shape) > 2:
        meer_data = meer_data[0, 0, :, :]  # meer data
    if meer_header["CTYPE3"] == "FREQ":
        frequency = meer_header["CRVAL3"] * u.Hz
    elif meer_header["CTYPE4"] == "FREQ":
        frequency = meer_header["CRVAL4"] * u.Hz
    else:
        frequency = ""
    if band == "":
        try:
            band = meer_header["BAND"]
        except BaseException:
            band = ""
    try:
        pixel_unit = meer_header["BUNIT"]
    except BaseException:
        pixel_nuit = ""
    obstime = Time(meer_header["date-obs"])
    meerpos = EarthLocation(
        lat=MEERLAT * u.deg, lon=MEERLON * u.deg, height=MEERALT * u.m
    )
    # Converting into GCRS coordinate
    meer_gcrs = SkyCoord(meerpos.get_gcrs(obstime))
    reference_coord = SkyCoord(
        meer_header["crval1"] * u.Unit(meer_header["cunit1"]),
        meer_header["crval2"] * u.Unit(meer_header["cunit2"]),
        frame="gcrs",
        obstime=obstime,
        obsgeoloc=meer_gcrs.cartesian,
        obsgeovel=meer_gcrs.velocity.to_cartesian(),
        distance=meer_gcrs.hcrs.distance,
    )
    reference_coord_arcsec = reference_coord.transform_to(
        frames.Helioprojective(observer=meer_gcrs)
    )
    cdelt1 = (np.abs(meer_header["cdelt1"]) * u.deg).to(u.arcsec)
    cdelt2 = (np.abs(meer_header["cdelt2"]) * u.deg).to(u.arcsec)
    P1 = sun.P(obstime)  # Relative rotation angle
    new_meer_header = make_fitswcs_header(
        meer_data,
        reference_coord_arcsec,
        reference_pixel=u.Quantity(
            [meer_header["crpix1"] - 1, meer_header["crpix2"] - 1] * u.pixel
        ),
        scale=u.Quantity([cdelt1, cdelt2] * u.arcsec / u.pix),
        rotation_angle=-P1,
        wavelength=frequency.to(u.MHz).round(2),
        observatory="MeerKAT",
    )
    if do_sharpen:
        blurred = gaussian_filter(meer_data, sigma=10)
        meer_data = meer_data + (meer_data - blurred)
    meer_map = Map(meer_data, new_meer_header)
    meer_map_rotate = meer_map.rotate()
    return meer_map_rotate


def save_in_hpc(fits_image, outdir="", xlim=[-1600, 1600], ylim=[-1600, 1600]):
    """
    Save solar image in helioprojective coordinates

    Parameters
    ----------
    fits_image : str
        FITS image name
    outdir : str, optional
        Output directory
    xlim : list
        X axis limit in arcsecond
    ylim : list
        Y axis limit in arcsecond

    Returns
    -------
    str
        FITS image in helioprojective coordinate
    """
    logging.getLogger("sunpy").setLevel(logging.ERROR)
    fits_header = fits.getheader(fits_image)
    meermap = get_meermap(fits_image)
    if len(xlim) == 2 and len(ylim) == 2:
        top_right = SkyCoord(
            xlim[1] * u.arcsec, ylim[1] * u.arcsec, frame=meermap.coordinate_frame
        )
        bottom_left = SkyCoord(
            xlim[0] * u.arcsec, ylim[0] * u.arcsec, frame=meermap.coordinate_frame
        )
        meermap = meermap.submap(bottom_left, top_right=top_right)
    if outdir == "":
        outdir = os.path.dirname(os.path.abspath(fits_image))
    outfile = f"{outdir}/{os.path.basename(fits_image).split('.fits')[0]}_HPC.fits"
    if os.path.exists(outfile):
        os.system(f"rm -rf {outfile}")
    meermap.save(outfile, filetype="fits")
    data = fits.getdata(outfile)
    data = data[np.newaxis, np.newaxis, ...]
    hpc_header = fits.getheader(outfile)
    for key in [
        "NAXIS",
        "NAXIS3",
        "NAXIS4",
        "BUNIT",
        "CTYPE3",
        "CRPIX3",
        "CRVAL3",
        "CDELT3",
        "CUNIT3",
        "CTYPE4",
        "CRPIX4",
        "CRVAL4",
        "CDELT4",
        "CUNIT4",
        "AUTHOR",
        "PIPELINE",
        "BAND",
        "MAX",
        "MIN",
        "RMS",
        "SUM",
        "MEAN",
        "MEDIAN",
        "RMSDYN",
        "MIMADYN",
    ]:
        if key in fits_header:
            hpc_header[key] = fits_header[key]
    fits.writeto(outfile, data=data, header=hpc_header, overwrite=True)
    return outfile


def plot_in_hpc(
    fits_image,
    draw_limb=False,
    extensions=["png"],
    outdirs=[],
    plot_range=[],
    power=0.5,
    xlim=[-1600, 1600],
    ylim=[-1600, 1600],
    contour_levels=[],
    band="",
    showgui=False,
):
    """
    Function to convert MeerKAT image into Helioprojective co-ordinate

    Parameters
    ----------
    fits_image : str
        Name of the fits image
    draw_limb : bool, optional
        Draw solar limb or not
    extensions : list, optional
        Output file extensions
    outdirs : list, optional
        Output directories for each extensions
    plot_range : list, optional
        Plot range
    power : float, optional
        Power stretch
    xlim : list
        X axis limit in arcsecond
    ylim : list
        Y axis limit in arcsecond
    contour_levels : list, optional
        Contour levels in fraction of peak, both positive and negative values allowed
    band : str, optional
        Band name
    showgui : bool, optional
        Show GUI

    Returns
    -------
    outfiles
        Saved plot file names
    sunpy.Map
        MeerKAT image in helioprojective co-ordinate
    """
    import matplotlib.ticker as ticker
    from matplotlib.patches import Ellipse, Rectangle
    from matplotlib.colors import ListedColormap
    from matplotlib import cm
    from sunpy.coordinates import sun

    logging.getLogger("sunpy").setLevel(logging.ERROR)
    if not showgui:
        matplotlib.use("Agg")
    else:
        matplotlib.use("TkAgg")
    matplotlib.rcParams.update({"font.size": 12})
    fits_image = fits_image.rstrip("/")
    meer_header = fits.getheader(fits_image)  # Opening MeerKAT fits file
    if meer_header["CTYPE3"] == "FREQ":
        frequency = meer_header["CRVAL3"] * u.Hz
    elif meer_header["CTYPE4"] == "FREQ":
        frequency = meer_header["CRVAL4"] * u.Hz
    else:
        frequency = ""
    if band == "":
        try:
            band = meer_header["BAND"]
        except BaseException:
            band = ""
    try:
        pixel_unit = meer_header["BUNIT"]
    except BaseException:
        pixel_nuit = ""
    obstime = Time(meer_header["date-obs"])
    meer_map_rotate = get_meermap(fits_image, band=band)
    top_right = SkyCoord(
        xlim[1] * u.arcsec, ylim[1] * u.arcsec, frame=meer_map_rotate.coordinate_frame
    )
    bottom_left = SkyCoord(
        xlim[0] * u.arcsec, ylim[0] * u.arcsec, frame=meer_map_rotate.coordinate_frame
    )
    cropped_map = meer_map_rotate.submap(bottom_left, top_right=top_right)
    meer_data = cropped_map.data
    if len(plot_range) < 2:
        norm = ImageNormalize(
            meer_data,
            vmin=0.03 * np.nanmax(meer_data),
            vmax=0.99 * np.nanmax(meer_data),
            stretch=PowerStretch(power),
        )
    else:
        norm = ImageNormalize(
            meer_data,
            vmin=np.nanmin(plot_range),
            vmax=np.nanmax(plot_range),
            stretch=PowerStretch(power),
        )
    if band == "U":
        cmap = "inferno"
        pos_color = "white"
        neg_color = "cyan"
    elif band == "L":
        pos_color = "hotpink"
        neg_color = "yellow"
        if "YlGnBu_inferno" not in plt.colormaps():
            # Sample YlGnBu_r colormap with 256 colors
            cmap_ylgnbu = cm.get_cmap("YlGnBu_r", 256)
            colors = cmap_ylgnbu(np.linspace(0, 1, 256))
            # Create perceptually linear spacing using inferno luminance
            cmap_inferno = cm.get_cmap("inferno", 256)
            # Sort YlGnBu colors by the inferred brightness from inferno
            luminance_ranks = np.argsort(
                np.mean(cmap_inferno(np.linspace(0, 1, 256))[:, :3], axis=1)
            )
            colors_uniform = colors[luminance_ranks]
            # New perceptual-YlGnBu-inspired colormap
            YlGnBu_inferno = ListedColormap(colors_uniform, name="YlGnBu_inferno")
            plt.colormaps.register(name="YlGnBu_inferno", cmap=YlGnBu_inferno)
        cmap = "YlGnBu_inferno"
    else:
        cmap = "cubehelix"
        pos_color = "cyan"
        neg_color = "gold"
    try:
        fig = plt.figure()
        ax = plt.subplot(projection=cropped_map)
        cropped_map.plot(norm=norm, cmap=cmap, axes=ax)
        if len(contour_levels) > 0:
            contour_levels = np.array(contour_levels)
            pos_cont = contour_levels[contour_levels >= 0]
            neg_cont = contour_levels[contour_levels < 0]
            if len(pos_cont) > 0:
                cropped_map.draw_contours(
                    np.sort(pos_cont) * np.nanmax(meer_data), colors=pos_color
                )
            if len(neg_cont) > 0:
                cropped_map.draw_contours(
                    np.sort(neg_cont) * np.nanmax(meer_data), colors=neg_color
                )
        ax.coords.grid(False)
        rgba_vmin = plt.get_cmap(cmap)(norm(norm.vmin))
        ax.set_facecolor(rgba_vmin)
        # Read synthesized beam from header
        try:
            bmaj = meer_header["BMAJ"] * u.deg.to(u.arcsec)  # in arcsec
            bmin = meer_header["BMIN"] * u.deg.to(u.arcsec)
            bpa = meer_header["BPA"] - sun.P(obstime).deg  # in degrees
        except KeyError:
            bmaj = bmin = bpa = None
        # Plot PSF ellipse in bottom-left if all values are present
        if bmaj and bmin and bpa is not None:
            # Coordinates where to place the beam (e.g., 5% above bottom-left
            # corner)
            x0, x1 = ax.get_xlim()
            y0, y1 = ax.get_ylim()

            beam_center = SkyCoord(
                x0 + 0.08 * (x1 - x0),
                y0 + 0.08 * (y1 - y0),
                unit=u.arcsec,
                frame=cropped_map.coordinate_frame,
            )

            # Add ellipse patch
            beam_ellipse = Ellipse(
                (beam_center.Tx.value, beam_center.Ty.value),  # center in arcsec
                width=bmin,
                height=bmaj,
                angle=bpa,
                edgecolor="white",
                facecolor="white",
                lw=1,
            )
            ax.add_patch(beam_ellipse)
            # Draw square box around the ellipse
            box_size = 100  # slightly bigger than beam
            rect = Rectangle(
                (
                    beam_center.Tx.value - box_size / 2,
                    beam_center.Ty.value - box_size / 2,
                ),
                width=box_size,
                height=box_size,
                edgecolor="white",
                facecolor="none",
                lw=1.2,
                linestyle="solid",
            )
            ax.add_patch(rect)
        if draw_limb:
            cropped_map.draw_limb()
        formatter = ticker.FuncFormatter(lambda x, _: f"{int(x):.0e}")
        cbar = plt.colorbar(format=formatter)
        # Optional: set max 5 ticks to prevent clutter
        cbar.locator = ticker.MaxNLocator(nbins=5)
        cbar.update_ticks()
        if pixel_unit == "K":
            cbar.set_label("Brightness temperature (K)")
        elif pixel_unit == "JY/BEAM":
            cbar.set_label("Flux density (Jy/beam)")
        fig.tight_layout()
        output_image_list = []
        for i in range(len(extensions)):
            ext = extensions[i]
            try:
                outdir = outdirs[i]
            except BaseException:
                outdir = os.path.dirname(os.path.abspath(fits_image))
            if len(contour_levels) > 0:
                output_image = (
                    outdir
                    + "/"
                    + os.path.basename(fits_image).split(".fits")[0]
                    + f"_contour.{ext}"
                )
            else:
                output_image = (
                    outdir
                    + "/"
                    + os.path.basename(fits_image).split(".fits")[0]
                    + f".{ext}"
                )
            output_image_list.append(output_image)
        for output_image in output_image_list:
            fig.savefig(output_image)
        if showgui:
            plt.show()
        plt.close(fig)
    except Exception:
        traceback.print_exc()
    finally:
        plt.close("all")
    return output_image_list, cropped_map


def get_suvi_map(obs_date, obs_time, workdir, wavelength=195):
    """
    Get GOES SUVI map

    Parameters
    ----------
    obs_date : str
        Observation date in yyyy-mm-dd format
    obs_time : str
        Observation time in hh:mm format
    workdir : str
        Work directory
    wavelength : float, optional
        Wavelength, options: 94, 131, 171, 195, 284, 304 Å

    Returns
    -------
    sunpy.map
        Sunpy SUVIMap
    """
    logging.getLogger("sunpy").setLevel(logging.ERROR)
    warnings.filterwarnings(
        "ignore",
        message="This download has been started in a thread which is not the main thread",
    )
    os.makedirs(workdir, exist_ok=True)
    start_time = dt.fromisoformat(f"{obs_date}T{obs_time}")
    t_start = (start_time - timedelta(minutes=2)).strftime("%Y-%m-%dT%H:%M")
    t_end = (start_time + timedelta(minutes=2)).strftime("%Y-%m-%dT%H:%M")
    time = a.Time(t_start, t_end)
    instrument = a.Instrument("suvi")
    wavelength = a.Wavelength(wavelength * u.angstrom)
    results = Fido.search(time, instrument, wavelength, a.Level(2))
    downloaded_files = Fido.fetch(results, path=workdir, progress=False)
    obs_times = []
    for image in downloaded_files:
        suvimap = Map(image)
        dateobs = suvimap.meta["date-obs"].split(".")[0]
        obs_times.append(dateobs)
    times_dt = [dt.strptime(t, "%Y-%m-%dT%H:%M:%S") for t in obs_times]
    closest_time = min(times_dt, key=lambda t: abs(t - start_time))
    pos = times_dt.index(closest_time)
    closest_time_str = closest_time.strftime("%Y-%m-%dT%H:%M")
    final_image = downloaded_files[pos]
    suvi_map = Map(final_image)
    for f in downloaded_files:
        os.system(f"rm -rf {f}")
    return suvi_map


def enhance_offlimb(sunpy_map, do_sharpen=True):
    """
    Enhance off-disk emission

    Parameters
    ----------
    sunpy_map : sunpy.map
        Sunpy map
    do_sharpen : bool, optional
        Sharpen images

    Returns
    -------
    sunpy.map
        Off-disk enhanced emission
    """
    from scipy.ndimage import gaussian_filter
    from sunpy.map.maputils import all_coordinates_from_map

    logging.getLogger("sunpy").setLevel(logging.ERROR)
    hpc_coords = all_coordinates_from_map(sunpy_map)
    r = np.sqrt(hpc_coords.Tx**2 + hpc_coords.Ty**2) / sunpy_map.rsun_obs
    rsun_step_size = 0.01
    rsun_array = np.arange(1, r.max(), rsun_step_size)
    y = np.array(
        [
            sunpy_map.data[(r > this_r) * (r < this_r + rsun_step_size)].mean()
            for this_r in rsun_array
        ]
    )
    pos = np.where(y < 10e-3)[0][0]
    r_lim = round(rsun_array[pos], 2)
    params = np.polyfit(
        rsun_array[rsun_array < r_lim], np.log(y[rsun_array < r_lim]), 1
    )
    scale_factor = np.exp((r - 1) * -params[0])
    scale_factor[r < 1] = 1
    if do_sharpen:
        blurred = gaussian_filter(sunpy_map.data, sigma=3)
        data = sunpy_map.data + (sunpy_map.data - blurred)
    else:
        data = sunpy_map.data
    scaled_map = Map(data * scale_factor, sunpy_map.meta)
    scaled_map.plot_settings["norm"] = ImageNormalize(stretch=LogStretch(10))
    return scaled_map


def make_meer_overlay(
    meerkat_image,
    suvi_wavelength=195,
    plot_file_prefix=None,
    plot_meer_colormap=True,
    enhance_offdisk=True,
    contour_levels=[0.05, 0.1, 0.2, 0.4, 0.6, 0.8],
    do_sharpen_suvi=True,
    xlim=[-1600, 1600],
    ylim=[-1600, 1600],
    extensions=["png"],
    outdirs=[],
    ncpu=-1,
    showgui=False,
    verbose=False,
):
    """
    Make overlay of MeerKAT image on GOES SUVI image

    Parameters
    ----------
    meerkat_image : str
        MeerKAT image
    suvi_wavelength : float, optional
        GOES SUVI wavelength, options: 94, 131, 171, 195, 284, 304 Å
    plot_file_prefix : str, optional
        Plot file prefix name
    plot_meer_colormap : bool, optional
        Plot MeerKAT map colormap
    enhance_offdisk : bool, optional
        Enhance off-disk emission
    contour_levels : list, optional
        Contour levels in fraction of peak
    do_sharpen_suvi : bool, optional
        Do sharpen SUVI images
    xlim : list, optional
        X-axis limit in arcsec
    tlim : list, optional
        Y-axis limit in arcsec
    extensions : list, optional
        Image file extensions
    outdirs : list, optional
        Output directories for each extensions
    ncpu : int, optional
        Number of CPUs to use
    showgui : bool, optional
        Show GUI
    verbose: bool, optinal
        Verbose output

    Returns
    -------
    list
        Plot file names
    """
    import matplotlib
    import matplotlib.ticker as ticker
    import matplotlib.pyplot as plt
    from sunpy.coordinates import SphericalScreen
    from matplotlib.colors import ListedColormap
    from matplotlib import cm
    from sunpy.map import make_fitswcs_header

    logging.getLogger("sunpy").setLevel(logging.ERROR)
    logging.getLogger("reproject.common").setLevel(logging.WARNING)

    @delayed
    def reproject_map(smap, target_header):
        with SphericalScreen(smap.observer_coordinate):
            return smap.reproject_to(target_header)

    if showgui:
        matplotlib.use("TkAgg")
    else:
        matplotlib.use("Agg")
    workdir = os.path.dirname(os.path.abspath(meerkat_image))
    meermap = get_meermap(meerkat_image)
    obs_datetime = fits.getheader(meerkat_image)["DATE-OBS"]
    obs_date = obs_datetime.split("T")[0]
    obs_time = ":".join(obs_datetime.split("T")[-1].split(":")[:2])
    suvi_map = get_suvi_map(obs_date, obs_time, workdir, wavelength=suvi_wavelength)
    if enhance_offdisk:
        suvi_map = enhance_offlimb(suvi_map, do_sharpen=do_sharpen_suvi)
    projected_coord = SkyCoord(
        0 * u.arcsec,
        0 * u.arcsec,
        obstime=suvi_map.observer_coordinate.obstime,
        frame="helioprojective",
        observer=suvi_map.observer_coordinate,
        rsun=suvi_map.coordinate_frame.rsun,
    )
    projected_header = make_fitswcs_header(
        suvi_map.data.shape,
        projected_coord,
        scale=u.Quantity(suvi_map.scale),
        instrument=suvi_map.instrument,
        wavelength=suvi_map.wavelength,
    )
    reprojected = [
        reproject_map(meermap, projected_header),
        reproject_map(suvi_map, projected_header),
    ]
    if ncpu < 1:
        ncpu = 1
    pool = ThreadPool(processes=ncpu)
    with dask.config.set(pool=pool):
        meer_reprojected, suvi_reprojected = compute(*reprojected, scheduler="threads")
    meertime = meermap.meta["date-obs"].split(".")[0]
    suvitime = suvi_map.meta["date-obs"].split(".")[0]
    try:
        if plot_meer_colormap and len(contour_levels) > 0:
            matplotlib.rcParams.update({"font.size": 18})
            fig = plt.figure(figsize=(16, 8))
            ax_colormap = fig.add_subplot(1, 2, 1, projection=suvi_reprojected)
            ax_contour = fig.add_subplot(1, 2, 2, projection=suvi_reprojected)
        elif plot_meer_colormap:
            matplotlib.rcParams.update({"font.size": 14})
            fig = plt.figure(figsize=(10, 8))
            ax_colormap = fig.add_subplot(projection=suvi_reprojected)
        elif len(contour_levels) > 0:
            matplotlib.rcParams.update({"font.size": 14})
            fig = plt.figure(figsize=(10, 8))
            ax_contour = fig.add_subplot(projection=suvi_reprojected)
        else:
            print("No overlay is plotting.")
            return

        title = f"SUVI time: {suvitime}\n MeerKAT time: {meertime}"
        if "transparent_inferno" not in plt.colormaps():
            cmap = cm.get_cmap("inferno", 256)
            colors = cmap(np.linspace(0, 1, 256))
            x = np.linspace(0, 1, 256)
            alpha = 0.8 * (1 - np.exp(-3 * x))
            colors[:, -1] = alpha  # Update the alpha channel
            transparent_inferno = ListedColormap(colors)
            plt.colormaps.register(name="transparent_inferno", cmap=transparent_inferno)
        if plot_meer_colormap and len(contour_levels) > 0:
            suptitle = title.replace("\n", ",")
            title = ""
            fig.suptitle(suptitle)
        if plot_meer_colormap:
            z = 0
            suvi_reprojected.plot(
                axes=ax_colormap,
                title=title,
                autoalign=True,
                clip_interval=(3, 99.9) * u.percent,
                zorder=z,
            )
            z += 1
            meer_reprojected.plot(
                axes=ax_colormap,
                title=title,
                clip_interval=(3, 99.9) * u.percent,
                cmap="transparent_inferno",
                zorder=z,
            )
        if len(contour_levels) > 0:
            z = 0
            suvi_reprojected.plot(
                axes=ax_contour,
                title=title,
                autoalign=True,
                clip_interval=(3, 99.9) * u.percent,
                zorder=z,
            )
            z += 1
            contour_levels = np.array(contour_levels) * np.nanmax(meer_reprojected.data)
            meer_reprojected.draw_contours(
                contour_levels, axes=ax_contour, cmap="YlGnBu", zorder=z
            )
            ax_contour.set_facecolor("black")

        if len(xlim) > 0:
            x_pix_limits = []
            for x in xlim:
                sky = SkyCoord(
                    x * u.arcsec, 0 * u.arcsec, frame=suvi_reprojected.coordinate_frame
                )
                x_pix = suvi_reprojected.world_to_pixel(sky)[0].value
                x_pix_limits.append(x_pix)
            if plot_meer_colormap and len(contour_levels) > 0:
                ax_colormap.set_xlim(x_pix_limits)
                ax_contour.set_xlim(x_pix_limits)
            elif plot_meer_colormap:
                ax_colormap.set_xlim(x_pix_limits)
            elif len(contour_levels) > 0:
                ax_contour.set_xlim(x_pix_limits)
        if len(ylim) > 0:
            y_pix_limits = []
            for y in ylim:
                sky = SkyCoord(
                    0 * u.arcsec, y * u.arcsec, frame=suvi_reprojected.coordinate_frame
                )
                y_pix = suvi_reprojected.world_to_pixel(sky)[1].value
                y_pix_limits.append(y_pix)
            if plot_meer_colormap and len(contour_levels) > 0:
                ax_colormap.set_ylim(y_pix_limits)
                ax_contour.set_ylim(y_pix_limits)
            elif plot_meer_colormap:
                ax_colormap.set_ylim(y_pix_limits)
            elif len(contour_levels) > 0:
                ax_contour.set_ylim(y_pix_limits)
        if plot_meer_colormap and len(contour_levels) > 0:
            ax_colormap.coords.grid(False)
            ax_contour.coords.grid(False)
        elif plot_meer_colormap:
            ax_colormap.coords.grid(False)
        elif len(contour_levels) > 0:
            ax_contour.coords.grid(False)
        fig.tight_layout()
        plot_file_list = []
        if verbose:
            print("#######################")
        if plot_file_prefix:
            for i in range(len(extensions)):
                ext = extensions[i]
                try:
                    savedir = outdirs[i]
                except BaseException:
                    savedir = workdir
                plot_file = f"{savedir}/{plot_file_prefix}.{ext}"
                plt.savefig(plot_file, bbox_inches="tight")
                if verbose:
                    print(f"Plot saved: {plot_file}")
                plot_file_list.append(plot_file)
            if verbose:
                print("#######################\n")
        else:
            plot_file = None
        if showgui:
            plt.show()
            plt.close(fig)
        else:
            plt.close(fig)
    except Exception:
        traceback.print_exc()
    finally:
        plt.close("all")
    return plot_file_list


##############################
# Extract dynamic spectrum
##############################
def make_ds_file_per_scan(msname, save_file, scan, datacolumn):
    """
    Extract dynamic spectrum from measurement set

    Parameters
    ----------
    msname : str
        Measurement set name
    save_file : str
        File name to save dynamic spectrum
    scan : int
        Scan number
    datacolumn : str
        Data column name

    Returns
    -------
    str
        Dynamic spectrum file
    """
    if os.path.exists(f"{save_file}.npy") == False:
        mstool = casamstool()
        try:
            all_data = []
            for ant in range(5):
                print(f"Extracting data for antenna :{ant}, scan: {scan}")
                mstool.open(msname)
                mstool.selectpolarization(["I"])
                mstool.select(
                    {"antenna1": ant, "antenna2": ant, "scan_number": int(scan)}
                )
                data_dic = mstool.getdata(datacolumn)
                mstool.close()
                if datacolumn == "CORRECTED_DATA":
                    data = np.abs(data_dic["corrected_data"][0, ...])
                else:
                    data = np.abs(data_dic["data"][0, ...])
                del data_dic
                m = np.nanmedian(data, axis=1)
                data = data / m[:, None]
                all_data.append(data)
                del data
        except Exception as e:
            print("Auto-corrrelations are not present. Using short baselines.")
            count = 0
            all_data = []
            while count <= 5:
                for i in range(5):
                    for j in range(5):
                        if i != j:
                            print(
                                f"Extracting data for antennas :{i} and {j}, scan: {scan}"
                            )
                            mstool.open(msname)
                            mstool.selectpolarization(["I"])
                            mstool.select(
                                {
                                    "antenna1": i,
                                    "antenna2": j,
                                    "scan_number": int(scan),
                                }
                            )
                            data_dic = mstool.getdata(datacolumn)
                            mstool.close()
                            if datacolumn == "CORRECTED_DATA":
                                data = np.abs(data_dic["corrected_data"][0, ...])
                            else:
                                data = np.abs(data_dic["data"][0, ...])
                            del data_dic
                            m = np.nanmedian(data, axis=1)
                            data = data / m[:, None]
                            all_data.append(data)
                            del data
                            count += 1
        finally:
            try:
                mstool.close()
            except:
                pass
        all_data = np.array(all_data)
        data = np.nanmedian(all_data, axis=0)
        bad_chans = get_bad_chans(msname)
        if bad_chans != "":
            bad_chans = bad_chans.replace("0:", "").split(";")
            for bad_chan in bad_chans:
                s = int(bad_chan.split("~")[0])
                e = int(bad_chan.split("~")[-1]) + 1
                data[s:e, :] = np.nan
        msmd = msmetadata()
        msmd.open(msname)
        freqs = msmd.chanfreqs(0, unit="MHz")
        times = msmd.timesforscans(int(scan))
        timestamps = [mjdsec_to_timestamp(mjdsec, str_format=0) for mjdsec in times]
        msmd.close()
        np.save(
            save_file,
            np.array([freqs, times, timestamps, data], dtype="object"),
        )
    if ".npy" in save_file:
        return save_file
    else:
        return f"{save_file}.npy"


def make_ds_plot(dsfiles, plot_file=None, showgui=False):
    """
    Make dynamic spectrum plot

    Parameters
    ----------
    dsfile : list
        DS files list
    plot_file : str, optional
        Plot file name to save the plot
    showgui : bool, optional
        Show GUI

    Returns
    -------
    str
        Plot name
    """
    from matplotlib.gridspec import GridSpec

    if showgui:
        matplotlib.use("TkAgg")
    else:
        matplotlib.use("Agg")
    matplotlib.rcParams.update({"font.size": 18})
    if type(dsfiles) == str:
        dsfiles = [dsfiles]
    for i, dsfile in enumerate(dsfiles):
        freqs_i, times_i, timestamps_i, data_i = np.load(dsfile, allow_pickle=True)
        if i == 0:
            freqs = freqs_i
            times = times_i
            timestamps = timestamps_i
            data = data_i
        else:
            gapsize = int(
                (np.nanmin(times_i) - np.nanmax(times)) / (times[1] - times[0])
            )
            if gapsize < 10:
                last_time_median = np.nanmedian(data[:, -1], axis=0)
                new_time_median = np.nanmedian(data_i[:, 0], axis=0)
                data_i = (data_i / new_time_median) * last_time_median
            # Insert vertical NaN gap (1 column wide)
            gap = np.full((data.shape[0], gapsize), np.nan)
            data = np.concatenate([data, gap, data_i], axis=1)
            # Insert dummy time and timestamp
            times = np.append(times, np.nan)
            timestamps = np.append(timestamps, "GAP")
            # Append new values
            times = np.append(times, times_i)
            timestamps = np.append(timestamps, timestamps_i)
            # (Optional) Check or merge freqs if needed — assuming same across files
    # Normalize by median bandshape
    median_bandshape = np.nanmedian(data, axis=-1)
    pos = np.where(np.isnan(median_bandshape) == False)[0]
    data /= median_bandshape[:, None]
    data = data[min(pos) : max(pos), :]
    freqs = freqs[min(pos) : max(pos)]
    temp_times = times[np.isnan(times) == False]
    maxtimepos = np.argmax(temp_times)
    mintimepos = np.argmin(temp_times)
    datestamp = f"{timestamps[mintimepos].split('T')[0]}"
    tstart = f"{timestamps[mintimepos].split('T')[0]} {':'.join(timestamps[mintimepos].split('T')[-1].split(':')[:2])}"
    tend = f"{timestamps[maxtimepos].split('T')[0]} {':'.join(timestamps[maxtimepos].split('T')[-1].split(':')[:2])}"
    print(f"Time range : {tstart}~{tend}")
    results = Fido.search(
        a.Time(tstart, tend), a.Instrument("XRS"), a.Resolution("avg1m")
    )
    files = Fido.fetch(results, path=os.path.dirname(dsfiles[0]), overwrite=False)
    goes_tseries = TimeSeries(files, concatenate=True)
    for goes_f in files:
        os.system(f"rm -rf {goes_f}")
    goes_tseries = goes_tseries.truncate(tstart, tend)
    timeseries = np.nanmean(data, axis=0)
    # Normalization
    data_std = np.nanstd(data)
    data_median = np.nanmedian(data)
    norm = ImageNormalize(
        data,
        stretch=LogStretch(1),
        vmin=0.99 * np.nanmin(data),
        vmax=0.99 * np.nanmax(data),
    )
    try:
        # Create figure and GridSpec layout
        fig = plt.figure(figsize=(18, 10))
        gs = GridSpec(
            nrows=3, ncols=2, width_ratios=[1, 0.03], height_ratios=[4, 1.5, 2]
        )
        # Axes
        ax_spec = fig.add_subplot(gs[0, 0])
        ax_ts = fig.add_subplot(gs[1, 0])
        ax_goes = fig.add_subplot(gs[2, 0])
        cax = fig.add_subplot(gs[:, 1])  # colorbar spans both rows
        # Plot dynamic spectrum
        im = ax_spec.imshow(
            data, aspect="auto", origin="lower", norm=norm, cmap="magma"
        )
        ax_spec.set_ylabel("Frequency (MHz)")
        ax_spec.set_xticklabels([])  # Remove x-axis labels from top plot
        # Y-ticks
        yticks = ax_spec.get_yticks()
        yticks = yticks[(yticks >= 0) & (yticks < len(freqs))]
        ax_spec.set_yticks(yticks)
        ax_spec.set_yticklabels([f"{freqs[int(i)]:.1f}" for i in yticks])
        # Plot time series
        ax_ts.plot(timeseries)
        ax_ts.set_xlim(0, len(timeseries) - 1)
        ax_ts.set_ylabel("Mean \n flux density")
        goes_tseries.plot(axes=ax_goes)
        goes_times = goes_tseries.time
        times_dt = goes_times.to_datetime()
        ax_goes.set_xlim(times_dt[0], times_dt[-1])
        ax_goes.set_ylabel(r"Flux ($\frac{W}{m^2}$)")
        ax_goes.legend(ncol=2, loc="upper right")
        ax_goes.set_title("GOES light curve", fontsize=14)
        ax_ts.set_title("MeerKAT light curve", fontsize=14)
        ax_spec.set_title("MeerKAT dynamic spectrum", fontsize=14)
        ax_goes.set_xlabel("Time (UTC)")
        # Format x-ticks
        ax_ts.set_xticks([])
        ax_ts.set_xticklabels([])
        # Colorbar
        cbar = fig.colorbar(im, cax=cax)
        cbar.set_label("Flux density (arb. unit)")
        plt.tight_layout()
        # Save or show
        if plot_file:
            plt.savefig(plot_file, bbox_inches="tight")
            print(f"Plot saved: {plot_file}")
        if showgui:
            plt.show()
            plt.close(fig)
        else:
            plt.close(fig)
    except Exception:
        traceback.print_exc()
    finally:
        plt.close("all")
    return plot_file


def plot_goes_full_timeseries(
    msname, workdir, plot_file_prefix=None, extension="png", showgui=False
):
    """
    Plot GOES full time series on the day of observation

    Parameters
    ----------
    msname : str
        Measurement set
    workdir : str
        Work directory
    plot_file_prefix : str, optional
        Plot file name prefix
    extension : str, optional
        Save file extension
    showgui : bool, optional
        Show GUI

    Returns
    -------
    str
        Plot file name
    """
    os.makedirs(workdir, exist_ok=True)
    if showgui:
        matplotlib.use("TkAgg")
    else:
        matplotlib.use("Agg")
    matplotlib.rcParams.update({"font.size": 14})
    scans, cal_scans, f_scans, g_scans, p_scans = get_cal_target_scans(msname)
    valid_scans = get_valid_scans(msname)
    filtered_scans = []
    for scan in scans:
        if scan in valid_scans:
            filtered_scans.append(scan)
    msmd = msmetadata()
    msmd.open(msname)
    tstart_mjd = min(msmd.timesforscan(int(min(filtered_scans))))
    tend_mjd = max(msmd.timesforscan(int(max(filtered_scans))))
    msmd.close()
    tstart = mjdsec_to_timestamp(tstart_mjd, str_format=2)
    tend = mjdsec_to_timestamp(tend_mjd, str_format=2)
    print(f"Time range: {tstart}~{tend}")
    results = Fido.search(
        a.Time(tstart, tend), a.Instrument("XRS"), a.Resolution("avg1m")
    )
    files = Fido.fetch(results, path=workdir, overwrite=False)
    goes_tseries = TimeSeries(files, concatenate=True)
    for f in files:
        os.system(f"rm -rf {f}")
    fig, ax = plt.subplots(figsize=(15, 5), constrained_layout=True)
    goes_tseries.plot(axes=ax)
    times = goes_tseries.time
    times_dt = times.to_datetime()
    ax.axvspan(tstart, tend, alpha=0.2)
    ax.set_xlim(times_dt[0], times_dt[-1])
    plt.tight_layout()
    # Save or show
    if plot_file_prefix:
        plot_file = f"{workdir}/{plot_file_prefix}.{extension}"
        plt.savefig(plot_file, bbox_inches="tight")
        print(f"Plot saved: {plot_file}")
    else:
        plot_file = None
    if showgui:
        plt.show()
        plt.close(fig)
        plt.close("all")
    else:
        plt.close(fig)
    return plot_file


def rename_meersolar_image(
    imagename,
    imagedir="",
    pol="",
    band="",
    attcal="NOINFO",
    cutout_rsun=2.5,
    make_overlay=True,
    make_plots=True,
):
    """
    Rename and move image to image directory

    Parameters
    ----------
    imagename : str
        Image name
    imagedir : str, optional
        Image directory (default given image directory)
    pol : str, optional
        Stokes parameters
    band : str, optional
        Observing band
    attcal : str, optional
        Solar attenuation calibrated or not
    cutout_rsun : float, optional
        Cutout in solar radii from center (default: 2.5 solar radii)
    make_overlay : bool, optional
        Make overlay on SUVI
    make_plots : bool, optional
        Make radio map plot in helioprojective coordinates

    Returns
    -------
    str
        New imagename with full path
    """
    imagename = imagename.rstrip("/")
    imagename = cutout_image(
        imagename, imagename, x_deg=(cutout_rsun * 2 * 16.0) / 60.0
    )
    header = fits.getheader(imagename)
    time = header["DATE-OBS"]
    astro_time = Time(time, scale="utc")
    sun_jpl = Horizons(id="10", location="500", epochs=astro_time.jd)
    eph = sun_jpl.ephemerides()
    sun_coords = SkyCoord(
        ra=eph["RA"][0] * u.deg, dec=eph["DEC"][0] * u.deg, frame="icrs"
    )
    maxval, minval, rms, total_val, mean_val, median_val, rms_dyn, minmax_dyn = (
        calc_solar_image_stat(imagename, disc_size=18)
    )
    with fits.open(imagename, mode="update") as hdul:
        hdr = hdul[0].header
        hdr["AUTHOR"] = "DevojyotiKansabanik,DeepanPatra"
        if band != "":
            hdr["BAND"] = band
        hdr["PIPELINE"] = "MeerSOLAR"
        hdr["CRVAL1"] = sun_coords.ra.deg
        hdr["CRVAL2"] = sun_coords.dec.deg
        hdr["MAX"] = maxval
        hdr["MIN"] = minval
        hdr["RMS"] = rms
        hdr["SUM"] = total_val
        hdr["MEAN"] = mean_val
        hdr["MEDIAN"] = median_val
        hdr["RMSDYN"] = rms_dyn
        hdr["MIMADYN"] = minmax_dyn
        hdr["ATTCAL"] = str(attcal)
    freq = round(header["CRVAL3"] / 10**6, 2)
    t_str = "".join(time.split("T")[0].split("-")) + (
        "".join(time.split("T")[-1].split(":"))
    )
    new_name = "time_" + t_str + "_freq_" + str(freq)
    if pol != "":
        new_name += "_pol_" + str(pol)
    if "MFS" in imagename:
        new_name += "_MFS"
    new_name = new_name + ".fits"
    if imagedir == "":
        imagedir = os.path.dirname(os.path.abspath(imagename))
    new_name = imagedir + "/" + new_name
    os.system("mv " + imagename + " " + new_name)
    hpcdir = f"{os.path.dirname(imagedir)}/images/hpcs"
    os.makedirs(hpcdir, exist_ok=True)
    save_in_hpc(new_name, outdir=hpcdir)
    if make_plots:
        try:
            pngdir = f"{os.path.dirname(imagedir)}/images/pngs"
            os.makedirs(pngdir, exist_ok=True)
            outimages, cropped_map = plot_in_hpc(
                new_name,
                draw_limb=True,
                extensions=["png"],
                outdirs=[pngdir],
            )
        except Exception:
            pass
    if make_overlay:
        try:
            overlay_pngdir = f"{os.path.dirname(imagedir)}/overlays_pngs"
            os.makedirs(overlay_pngdir, exist_ok=True)
            outimages = make_meer_overlay(
                new_name,
                plot_file_prefix=os.path.basename(new_name).split(".fits")[0]
                + "_suvi_meerkat_overlay",
                extensions=["png"],
                outdirs=[overlay_pngdir],
                verbose=False,
            )
        except Exception:
            pass
    return new_name


# Expose functions and classes
__all__ = [
    name
    for name, obj in globals().items()
    if (
        (isinstance(obj, types.FunctionType) or isinstance(obj, type))
        and obj.__module__ == __name__
    )
]
